"""JSON utilities."""

from __future__ import annotations

import json
from typing import Any

import numpy as np
import pandas as pd


class NumpyEncoder(json.JSONEncoder):
    """Custom JSONEncoder for numpy data types.

    Examples
    --------
    >>> import json
    >>> import numpy as np
    >>> from pycontrails.utils.json import NumpyEncoder

    >>> data = np.array([0, 1, 2, 3])
    >>> json.dumps(data, cls=NumpyEncoder)
    '[0, 1, 2, 3]'

    >>> data = np.datetime64(1234567890, "s")
    >>> json.dumps(data, cls=NumpyEncoder)
    '"2009-02-13T23:31:30"'

    Notes
    -----
    Adapted https://github.com/hmallen/numpyencoder/blob/master/numpyencoder/numpyencoder.py
    """

    def default(self, obj: Any) -> Any:
        """Encode numpy data types.

        This method overrides :meth:`default` on the JSONEncoder class.

        Parameters
        ----------
        obj : Any
            Object to encode.

        Returns
        -------
        Any
            Encoded object.
        """
        if isinstance(
            obj,
            (
                np.int_,
                np.intc,
                np.intp,
                np.int8,
                np.int16,
                np.int32,
                np.int64,
                np.uint8,
                np.uint16,
                np.uint32,
                np.uint64,
            ),
        ):
            return int(obj)

        if isinstance(obj, (np.float_, np.float16, np.float32, np.float64)):
            return float(obj)

        # TODO: this is not easily reversible - np.timedelta64(str(np.timedelta64(1, "h"))) raises
        if isinstance(obj, (np.timedelta64)):
            return str(obj)

        if isinstance(obj, (np.datetime64)):
            return str(obj)

        if isinstance(obj, (np.complex_, np.complex64, np.complex128)):
            return {"real": obj.real, "imag": obj.imag}

        if isinstance(obj, (np.ndarray,)):
            return obj.tolist()

        if isinstance(obj, (np.bool_)):
            return bool(obj)

        if isinstance(obj, (np.void)):
            return None

        if isinstance(obj, (pd.Series, pd.Index)):
            return obj.to_numpy().tolist()

        try:
            return json.JSONEncoder.default(self, obj)

        # last ditch attempt by looking for a to_json attribute
        except TypeError as e:
            try:
                return obj.to_json
            except AttributeError:
                raise TypeError from e


def dataframe_to_geojson_points(
    df: pd.DataFrame,
    properties: list[str] | None = None,
    filter_nan: bool | list[str] = False,
) -> dict[str, Any]:
    """Convert a pandas DataFrame to a GeoJSON-like dictionary.

    This function create a Python representation of a GeoJSON FeatureCollection with Point features
    based on a :class:`pandas.DataFrame` with geospatial coordinate columns.

    Parameters
    ----------
    df : pd.DataFrame
        Base dataframe.
        Must contain geospatial coordinate columns ["longitude", "latitude", "altitude", "time"]
    properties : list[str], optional
        Specify columns to include feature properties.
        By default, will use all column data that is not in the coordinate columns.
    filter_nan : bool | list[str], optional
        Filter out points with nan values in any columns, including coordinate columns.
        If `list of str` is input, only `filter_nan` columns will be used for filtering,
        allowing null values in the other columns.

    Returns
    -------
    dict[str, Any]
        Description

    Raises
    ------
    KeyError
        Raises if `properties` or `filter_nan` input contains a column label that does
        not exist in `df`
    """
    # required columns
    coord_cols = ["longitude", "latitude", "altitude", "time"]

    # get all properties if not defined
    if properties is None:
        properties = [c for c in df.columns if c not in coord_cols]
    elif [c for c in properties if c not in df.columns]:
        raise KeyError(
            f"{[c for c in properties if c not in df.columns]} do not exist in dataframe"
        )

    # downselect dataframe
    cols = ["longitude", "latitude", "altitude", "time"] + properties
    df = df[cols]

    # filter out coords with nan values, or filter just on "filter_nan" labels
    if isinstance(filter_nan, bool) and filter_nan:
        df = df[np.all(~np.isnan(df[cols]), axis=1)]
    elif isinstance(filter_nan, list):
        if [c for c in filter_nan if c not in df.columns]:
            raise KeyError(
                f"{[c for c in filter_nan if c not in df.columns]} do not exist in dataframe"
            )
        df = df[np.all(~np.isnan(df[filter_nan]), axis=1)]

    def row_to_feature(row: pd.Series) -> dict[str, str | dict[str, Any]]:
        point = [
            np.round(row.longitude, decimals=4) if not np.isnan(row.longitude) else None,
            np.round(row.latitude, decimals=4) if not np.isnan(row.latitude) else None,
            np.round(row.altitude, decimals=4) if not np.isnan(row.altitude) else None,
        ]
        # converting to int to allow JSON serialization
        properties = {"time": int(row.time.timestamp())}
        used_keys = ["time", "latitude", "longitude", "altitude"]
        unused_keys = [k for k in row.keys() if k not in used_keys]
        properties.update(
            {
                k: (
                    row[k]
                    if not isinstance(row[k], float)
                    or isinstance(row[k], float)
                    and not np.isnan(row[k])
                    else None
                )
                for k in unused_keys
            }
        )

        geometry = {"type": "Point", "coordinates": point}
        return {"type": "Feature", "geometry": geometry, "properties": properties}

    features = []
    df.apply(lambda row: features.append(row_to_feature(row)), axis=1)
    return {"type": "FeatureCollection", "features": features}



"""Lightweight data structures for vector paths."""

from __future__ import annotations

import hashlib
import json
import logging
import warnings
from typing import Any, Dict, Generator, Iterable, Iterator, Sequence, Type, TypeVar, overload

import numpy as np
import numpy.typing as npt
import pandas as pd
import pyproj
from overrides import overrides

from pycontrails.core import coordinates, interpolation
from pycontrails.core import met as met_module
from pycontrails.physics import units
from pycontrails.utils import json as json_module

logger = logging.getLogger(__name__)

#: Vector types
VectorDatasetType = TypeVar("VectorDatasetType", bound="VectorDataset")
GeoVectorDatasetType = TypeVar("GeoVectorDatasetType", bound="GeoVectorDataset")


class AttrDict(Dict[str, Any]):
    """Thin wrapper around dict to warn when setting a key that already exists."""

    def __setitem__(self, k: str, v: Any) -> None:
        """Warn when setting values that already contain values.

        Parameters
        ----------
        k : str
            Key
        v : Any
            Value
        """
        if k in self and self[k] is not None and self[k] is not v:
            warnings.warn(
                f"Overwriting attr key `{k}`. Use `.update({k}=...)` to suppress warning."
            )

        super().__setitem__(k, v)

    def setdefault(self, k: str, default: Any = None) -> Any:
        """Thin wrapper around ``dict.setdefault``.

        Overwrites value if value is None.

        Parameters
        ----------
        k : str
            Key
        default : Any, optional
            Default value for key ``k``

        Returns
        -------
        Any
            Value at ``k``
        """
        ret = self.get(k, None)
        if ret is not None:
            return ret

        self[k] = default
        return default


class VectorDataDict(Dict[str, np.ndarray]):
    """Thin wrapper around ``dict[str, np.ndarray]`` to ensure consistency.

    Parameters
    ----------
    data : dict[str, np.ndarray], optional
        Dictionary input
    **kwargs : np.ndarray
        Keyword arguments, override values in ``data``
    """

    __slots__ = "_size"

    #: Length of the data
    _size: int

    def __init__(self, data: dict[str, np.ndarray] | None = None, **kwargs: np.ndarray) -> None:
        if data is None:
            data = {}

        super().__init__(data, **kwargs)

        # validate any arrays, first one defines size()
        for arr in self.values():
            self._validate_array(arr)

    def __setitem__(self, k: str, v: npt.ArrayLike) -> None:
        """Set new key-value pair to instance and warn when overwriting existing key.

        This method casts ``v`` to a ``np.ndarray`` and ensures that the array size is
        consistent with the instance.

        Parameters
        ----------
        k : str
            Key
        v : np.ndarray
            Values

        See Also
        --------
        :meth:`update`
        """
        v = np.asarray(v)  # asarray does NOT copy
        self._validate_array(v)

        if k in self and len(self[k]) and self[k] is not v:
            warnings.warn(
                f"Overwriting data in key `{k}`. Use `.update({k}=...)` to suppress warning."
            )

        super().__setitem__(k, v)

    def __delitem__(self, k: str) -> None:
        super().__delitem__(k)

        # if not data keys left, set size to 0
        if not len(self):
            del self._size

    def setdefault(self, k: str, default: npt.ArrayLike | None = None) -> np.ndarray:
        """Thin wrapper around ``dict.setdefault``.

        The main purpose of overriding is to run :meth:`_validate_array()` on set.

        Parameters
        ----------
        k : str
            Key
        default : npt.ArrayLike, optional
            Default value for key ``k``

        Returns
        -------
        Any
            Value at ``k``
        """
        ret = self.get(k, None)
        if ret is not None:
            return ret

        if default is None:
            default = np.array([])

        self[k] = default
        return self[k]

    def update(  # type: ignore[override]
        self, other: dict[str, npt.ArrayLike] | None = None, **kwargs: npt.ArrayLike
    ) -> None:
        """Update values without warning if overwriting.

        This method casts values in  ``other`` to np.ndarray arrays and ensures that the
        array sizes are consistent with the instance.

        Parameters
        ----------
        other : dict[str, npt.ArrayLike] | None, optional
            Fields to update as dict
        **kwargs : npt.ArrayLike
            Fields to update as kwargs
        """
        other = other or {}
        other_arrs = {k: np.asarray(v) for k, v in other.items()}
        for arr in other_arrs.values():
            self._validate_array(arr)

        super().update(other_arrs)

        # validate any kwarg arrays
        kwargs_arr = {k: np.asarray(v) for k, v in kwargs.items()}
        for arr in kwargs_arr.values():
            self._validate_array(arr)

        super().update(kwargs_arr)

    def _validate_array(self, arr: np.ndarray) -> None:
        """Ensure that `arr` is compatible with instance.

        Set attribute `_size` if it has not yet been defined.

        Parameters
        ----------
        arr : np.ndarray
            Array to validate

        Raises
        ------
        ValueError
            If `arr` is not compatible with instance.
        """
        if arr.ndim != 1:
            raise ValueError("All np.arrays must have dimension 1.")
        if getattr(self, "_size", 0) != 0:
            if arr.size != self._size:
                raise ValueError(f"Incompatible array sizes: {arr.size} and {self._size}.")
        else:
            self._size = arr.size


def _empty_vector_dict(keys: Iterable[str]) -> VectorDataDict:
    """Create instance of VectorDataDict with variables defined by `keys` and size 0.

    Parameters
    ----------
    keys : Iterable[str]
        Keys to include in empty VectorDataset instance.

    Returns
    -------
    VectorDataDict
        Empty :class:`VectorDataDict` instance.
    """
    keys = keys or set()
    data = VectorDataDict({key: np.array([]) for key in keys})

    # The default dtype is float64
    # Time is special and should have a non-default dtype of datetime64[ns]
    if "time" in data:
        data.update(time=np.array([], dtype="datetime64[ns]"))

    return data


class VectorDataset:
    """Base class to hold 1D arrays of consistent size.

    Parameters
    ----------
    data : dict[str, npt.ArrayLike] | pd.DataFrame | VectorDataDict | VectorDataset | None, optional
        Initial data, by default None
    attrs : dict[str, Any] | AttrDict, optional
        Dictionary of attributes, by default None
    copy : bool, optional
        Copy data on class creation, by default True
    **attrs_kwargs : Any
        Additional attributes passed as keyword arguments

    Raises
    ------
    ValueError
        If "time" variable cannot be converted to numpy array.
    """

    __slots__ = ("data", "attrs")

    #: Vector data with labels as keys and np.ndarrays as values
    data: VectorDataDict

    #: Generic dataset attributes
    attrs: AttrDict

    def __init__(
        self,
        data: dict[str, npt.ArrayLike]
        | pd.DataFrame
        | VectorDataDict
        | VectorDataset
        | None = None,
        attrs: dict[str, Any] | AttrDict | None = None,
        copy: bool = True,
        **attrs_kwargs: Any,
    ) -> None:
        # Casting from one VectorDataset type to another
        # e.g., flight = Flight(...); vector = VectorDataset(flight)
        if isinstance(data, VectorDataset):
            attrs = {**data.attrs, **(attrs or {})}
            data = data.data

        if data is None:
            self.data = VectorDataDict()

        elif isinstance(data, pd.DataFrame):
            # Take extra caution with a time column

            if "time" in data:
                time = data["time"]

                if not hasattr(time, "dt"):
                    # If the time column is a string, we try to convert it to a datetime
                    # If it fails (for example, a unix integer time), we raise an error
                    # and let the user figure it out.
                    try:
                        time = pd.to_datetime(time)
                    except ValueError:
                        raise ValueError(
                            "The 'time' field must hold datetime-like values. "
                            'Try data["time"] = pd.to_datetime(data["time"], unit=...) '
                            "with the appropriate unit."
                        )

                # If the time column contains a timezone, the call to `to_numpy`
                # will convert it to an array of object. We do not want this, so
                # we raise an error in this case. Timezone issues are complicated,
                # and so it is better for the user to handle them rather than try
                # to address them here.
                if time.dt.tz is not None:
                    raise ValueError(
                        "The 'time' field must be timezone naive. "
                        "This can be achieved with: "
                        'data["time"] = data["time"].dt.tz_localize(None)'
                    )

                data = {col: ser.to_numpy(copy=copy) for col, ser in data.items() if col != "time"}
                data["time"] = time.to_numpy(copy=copy)
            else:
                data = {col: ser.to_numpy(copy=copy) for col, ser in data.items()}

            self.data = VectorDataDict(data)

        elif isinstance(data, VectorDataDict) and not copy:
            self.data = data

        # For anything else, we assume it is a dictionary of array-like and attach it
        # This doesn't quite work for casting one VectorDataset to another subclass
        # (we don't support .items() here), but we could easily accommodate that if
        # needed
        else:
            self.data = VectorDataDict({k: np.array(v, copy=copy) for k, v in data.items()})

        # set attributes
        if attrs is None:
            self.attrs = AttrDict()

        elif isinstance(attrs, AttrDict) and not copy:
            self.attrs = attrs

        #  shallow copy if dict
        else:
            self.attrs = AttrDict(attrs.copy())

        # update with kwargs
        self.attrs.update(attrs_kwargs)

    # ------------
    # dict-like methods
    # ------------
    def __getitem__(self, key: str) -> np.ndarray:
        """Get values from :attr:`data`.

        Parameters
        ----------
        key : str
            Key to get from :attr:`data`

        Returns
        -------
        np.ndarray
            Values at :attr:`data[key]`
        """
        return self.data[key]

    def get(self, key: str, default_value: Any = None) -> Any:
        """Get values from :attr:`data` with default_value if `key` not in `data`.

        Parameters
        ----------
        key : str
            Key to get from :attr:`data`
        default_value : Any, optional
            Return `default_value` if `key` not in :attr:`data`, by default `None`

        Returns
        -------
        Any
            Values at :attr:`data[key]` or `default_value`
        """
        return self.data.get(key, default_value)

    def __setitem__(self, key: str, values: npt.ArrayLike) -> None:
        """Set values at key `key` on :attr:`data`.

        Parameters
        ----------
        key : str
            Key name in :attr:`data`
        values : npt.ArrayLike
            Values to set to :attr:`data`. Array size must be compatible with existing data.
        """
        self.data[key] = values

    def __delitem__(self, key: str) -> None:
        """Delete values at key `key` on :attr:`data`.

        Parameters
        ----------
        key : str
            Key name in :attr:`data`
        """
        del self.data[key]

    def __iter__(self) -> Iterator[str]:
        """Iterate over keys in :attr:`data`.

        Returns
        -------
        Iterator[str]
            Keys in :attr:`data`
        """
        return iter(self.data)

    def __contains__(self, key: str) -> bool:
        """Check if key `key` is in :attr:`data`.

        Parameters
        ----------
        key : str
            Key to check

        Returns
        -------
        bool
            True if `key` is in :attr:`data`, False otherwise
        """
        return key in self.data

    def update(
        self,
        other: dict[str, npt.ArrayLike] | None = None,
        **kwargs: npt.ArrayLike,
    ) -> None:
        """Update values in  :attr:`data` dict without warning if overwriting.

        Parameters
        ----------
        other : dict[str, npt.ArrayLike] | None, optional
            Fields to update as dict
        **kwargs : npt.ArrayLike
            Fields to update as kwargs
        """
        self.data.update(other, **kwargs)

    def setdefault(self, key: str, default: npt.ArrayLike | None = None) -> np.ndarray:
        """Shortcut to :attr:`data.setdefault`.

        Parameters
        ----------
        key : str
            Key in :attr:`data` dict.
        default : npt.ArrayLike, optional
            Values to use as default, if key is not defined

        Returns
        -------
        np.ndarray
            Values at ``key``
        """
        return self.data.setdefault(key, default)

    def get_data_or_attr(self, key: str) -> Any:
        """Get value from :attr:`data` or :attr:`attrs`.

        This method first checks if ``key`` is in :attr:`data` and returns the value if so.
        If ``key`` is not in :attr:`data`, then this method checks if ``key`` is in :attr:`attrs`
        and returns the value if so. If ``key`` is not in :attr:`data` or :attr:`attrs`,
        then a ``KeyError`` is raised.

        Parameters
        ----------
        key : str
            Key to get from :attr:`data` or :attr:`attrs`

        Returns
        -------
        Any
            Value at :attr:`data[key]` or :attr:`attrs[key]`

        Examples
        --------
        >>> vector = VectorDataset({"a": [1, 2, 3]}, attrs={"b": 4})
        >>> vector.get_data_or_attr("a")
        array([1, 2, 3])

        >>> vector.get_data_or_attr("b")
        4

        >>> vector.get_data_or_attr("c")
        Traceback (most recent call last):
        ...
        KeyError: 'c'

        """
        try:
            return self.data[key]
        except KeyError:
            return self.attrs[key]

    # ------------

    def __len__(self) -> int:
        """Length of each array in :attr:`data`.

        Returns
        -------
        int
            Length of each array in :attr:`data`
        """
        return self.size

    def _display_attrs(self) -> dict[str, str]:
        """Return properties used in `repr` constructions`.

        Returns
        -------
        dict[str, str]
            Properties used in :meth:`__repr__` and :meth:`_repr_html_`.
        """

        # Clip any attribute value that is too long
        def str_clip(v: Any) -> str:
            s = str(v)
            if len(s) < 80:
                return s
            return f"{s[:77]}..."

        return {k: str_clip(v) for k, v in self.attrs.items()}

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        n_attrs = len(self.attrs)
        n_keys = len(self.data)
        _repr = f"{class_name} [{n_keys} keys x {self.size} length, {n_attrs} attributes]"

        keys = list(self.data.keys())
        keys = keys[0:5] + ["..."] + keys[-1:] if len(keys) > 5 else keys
        _repr += f"\n\tKeys: {', '.join(keys)}"

        attrs = self._display_attrs()
        _repr += "\n\tAttributes:\n"
        _repr += "\n".join([f"\t{k:20}{v}" for k, v in attrs.items()])

        return _repr

    def _repr_html_(self) -> str:
        name = type(self).__name__
        n_attrs = len(self.attrs)
        n_keys = len(self.data)
        attrs = self._display_attrs()
        size = self.size

        title = f"<b>{name}</b> [{n_keys} keys x {size} length, {n_attrs} attributes]<br/ ><br/>"

        # matching pd.DataFrame styling
        header = '<tr style="border-bottom:1px solid silver"><th colspan="2">Attributes</th></tr>'
        rows = [f"<tr><td>{k}</td><td>{v}</td></tr>" for k, v in attrs.items()]
        table = f"<table>{header + ''.join(rows)}</table>"
        return title + table + self.dataframe._repr_html_()

    def __bool__(self) -> bool:
        """Check if :attr:`data` is nonempty..

        Returns
        -------
        bool
            True if non-empty values are set in :attr:`data`
        """
        return self.size > 0

    def __add__(self: VectorDatasetType, other: VectorDatasetType | None) -> VectorDatasetType:
        """Concatenate two compatible instances of VectorDataset.

        In this context, compatibility means that both have identical :attr:`data` keys.

        This operator behaves similarly to the ``__add__`` method on python lists.

        If self is an empty VectorDataset, return other. This is useful when
        calling :keyword:`sum` with an empty initial value.

        Parameters
        ----------
        other : VectorDatasetType
            Other values to concatenate

        Returns
        -------
        VectorDatasetType
            Concatenated values.

        Raises
        ------
        KeyError
            If `other` has different :attr:`data` keys than self.
        """
        # Short circuit: If other is empty or None, return self. The order here can matter.
        # We let self (so the left addend) take priority.
        if not other:
            return self
        if not self:
            return other

        return type(self).sum([self, other])

    @classmethod
    def sum(
        cls: Type[VectorDatasetType],
        vectors: Sequence[VectorDataset],
        infer_attrs: bool = True,
    ) -> VectorDatasetType:
        """Sum a list of :class:`VectorDataset` instances.

        Parameters
        ----------
        vectors : Sequence[VectorDataset]
            List of :class:`VectorDataset` instances to concatenate.
        infer_attrs : bool, optional
            If True, infer attributes from the first VectorDataset in the list.

        Returns
        -------
        VectorDataset
            Sum of all VectorDataset instances in ``vectors``.

        Raises
        ------
        KeyError
            If incompatible :attr:`data` keys are found among ``vectors``.

        Examples
        --------
        >>> from pycontrails import VectorDataset
        >>> v1 = VectorDataset({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> v2 = VectorDataset({"a": [7, 8, 9], "b": [10, 11, 12]})
        >>> v3 = VectorDataset({"a": [13, 14, 15], "b": [16, 17, 18]})
        >>> v = VectorDataset.sum([v1, v2, v3])
        >>> v.dataframe
            a   b
        0   1   4
        1   2   5
        2   3   6
        3   7  10
        4   8  11
        5   9  12
        6  13  16
        7  14  17
        8  15  18

        """
        if not vectors:
            return cls()
        keys = vectors[0].data.keys()
        for v in vectors[1:]:
            if v.data.keys() != keys:
                raise KeyError("Summands have incompatible keys.")

        def concat(key: str) -> np.ndarray:
            values = [v[key] for v in vectors]
            return np.concatenate(values)

        data = {key: concat(key) for key in keys}

        if infer_attrs:
            return cls(data, attrs=vectors[0].attrs)
        return cls(data)

    def __eq__(self: VectorDatasetType, other: object) -> bool:
        """Determine if two instances are equal.

        NaN values are considered equal in this comparison.

        Parameters
        ----------
        other : object
            VectorDatasetType to compare with

        Returns
        -------
        bool
            True if both instances have identical :attr:`data` and :attr:`attrs`.
        """
        if isinstance(other, VectorDataset):
            # assert attrs equal
            for key in self.attrs:
                if isinstance(self.attrs[key], np.ndarray):
                    # equal_nan not supported for non-numeric data
                    equal_nan = not np.issubdtype(self.attrs[key].dtype, "O")
                    try:
                        eq = np.array_equal(self.attrs[key], other.attrs[key], equal_nan=equal_nan)
                    except KeyError:
                        return False
                else:
                    eq = self.attrs[key] == other.attrs[key]

                if not eq:
                    return False

            # assert data equal
            for key in self:
                # equal_nan not supported for non-numeric data (e.g. strings)
                equal_nan = not np.issubdtype(self[key].dtype, "O")
                try:
                    eq = np.array_equal(self[key], other[key], equal_nan=equal_nan)
                except KeyError:
                    return False

                if not eq:
                    return False

            return True
        return False

    @property
    def size(self) -> int:
        """Length of each array in :attr:`data`.

        Returns
        -------
        int
            Length of each array in :attr:`data`.
        """
        return getattr(self.data, "_size", 0)

    @property
    def shape(self) -> tuple[int]:
        """Shape of each array in :attr:`data`.

        Returns
        -------
        tuple[int]
            Shape of each array in :attr:`data`.
        """
        return (self.size,)

    @property
    def dataframe(self) -> pd.DataFrame:
        """Shorthand property to access :meth:`to_dataframe` with `copy=False`.

        Returns
        -------
        pd.DataFrame
            Equivalent to the output from :meth:`to_dataframe()`
        """
        return self.to_dataframe(copy=False)

    @property
    def hash(self) -> str:
        """Generate a unique hash for this class instance.

        Returns
        -------
        str
            Unique hash for flight instance (sha1)
        """
        _hash = json.dumps(self.data, cls=json_module.NumpyEncoder)
        return hashlib.sha1(bytes(_hash, "utf-8")).hexdigest()

    # ------------
    # Utilities
    # ------------

    def copy(self: VectorDatasetType) -> VectorDatasetType:
        """Return a copy of this VectorDatasetType class.

        Returns
        -------
        VectorDatasetType
            Copy of class
        """
        return type(self)(data=self.data, attrs=self.attrs, copy=True)

    def select(self: VectorDataset, keys: list[str], copy: bool = True) -> VectorDataset:
        """Return new class instance only containing specified keys.

        Parameters
        ----------
        keys : list[str]
            A list of keys to filter by.
        copy : bool, optional
            Copy data on selection.
            Defaults to True.

        Returns
        -------
        VectorDataset
            VectorDataset containing only data associated to `keys`.
            Note that this method always returns a VectorDataset, even if
            the calling class is an implementation (i.e. Flight, Contrail)
        """
        data = {key: self[key] for key in keys}
        return VectorDataset(data=data, attrs=self.attrs, copy=copy)

    def filter(
        self: VectorDatasetType, mask: npt.NDArray[np.bool_], copy: bool = True
    ) -> VectorDatasetType:
        """Filter :attr:`data` according to a boolean array ``mask``.

        Entries corresponding to ``mask == True`` are kept.

        Parameters
        ----------
        mask : npt.NDArray[np.bool_]
            Boolean array with compatible shape.
        copy : bool, optional
            Copy data on filter. Defaults to True. See
            `numpy best practices <https://numpy.org/doc/stable/user/basics.indexing.html#slicing-and-striding>`_
            for insight into whether copy is appropriate.

        Returns
        -------
        VectorDatasetType
            Containing filtered data

        Raises
        ------
        TypeError
            If ``mask`` is not a boolean array.
        """  # noqa: E501
        self.data._validate_array(mask)
        if mask.dtype != bool:
            raise TypeError("Parameter `mask` must be a boolean array.")

        data = {key: value[mask] for key, value in self.data.items()}
        return type(self)(data=data, attrs=self.attrs, copy=copy)

    def sort(self: VectorDatasetType, by: str | list[str]) -> VectorDatasetType:
        """Sort data by key(s).

        This method always creates a copy of the data.

        Parameters
        ----------
        by : str | list[str]
            Key or list of keys to sort by.

        Returns
        -------
        VectorDatasetType
            Instance with sorted data.
        """
        return type(self)(data=self.dataframe.sort_values(by=by), attrs=self.attrs)

    def ensure_vars(self, vars: str | Sequence[str], raise_error: bool = True) -> bool:
        """Ensure variables exist in column of :attr:`data` or :attr:`attrs`.

        Parameters
        ----------
        vars : str | Sequence[str]
            A single string variable name or a sequence of string variable names.
        raise_error : bool, optional
            Raise KeyError if data does not contain variables.
            Defaults to True.

        Returns
        -------
        bool
            True if all variables exist.
            False otherwise.

        Raises
        ------
        KeyError
            Raises when dataset does not contain variable in ``vars``
        """
        if isinstance(vars, str):
            vars = (vars,)

        for v in vars:
            if v in self or v in self.attrs:
                continue
            if raise_error:
                raise KeyError(
                    f"{type(self).__name__} instance does not contain data or attr '{v}'"
                )
            return False

        return True

    def broadcast_attrs(self, keys: str | Iterable[str], overwrite: bool = False) -> None:
        """Attach values from `keys` in :attr:`attrs` onto :attr:`data`.

        If possible, use ``dtype = np.float32`` when broadcasting. If not possible,
        use whatever ``dtype`` is inferred from the data by :func:`numpy.full`.

        Parameters
        ----------
        keys : str | Iterable[str]
            Keys to broadcast
        overwrite : bool, optional
            If True, overwrite existing values in :attr:`data`. By default False.

        Raises
        ------
        KeyError
            Not all `keys` found in :attr:`attrs`.
        """
        if isinstance(keys, str):
            keys = [keys]
        else:
            keys = list(keys)  # necessary since we iterate over it twice

        # Validate everything up front to avoid partial broadcasting
        for key in keys:
            if key not in self.attrs:
                raise KeyError(f"{type(self)} does not contain attr `{key}`")

        # Do the broadcasting
        for attr in keys:
            if attr in self.data and not overwrite:
                warnings.warn(
                    f"Found duplicate key {attr} in attrs and data. "
                    "Set `overwrite=True` parameter to force overwrite."
                )
            else:
                scalar = self.attrs[attr]
                min_dtype = np.min_scalar_type(scalar)
                if np.can_cast(min_dtype, np.float32):
                    self.data.update({attr: np.full(self.size, scalar, dtype=np.float32)})
                else:
                    self.data.update({attr: np.full(self.size, scalar)})

    def broadcast_numeric_attrs(
        self, ignore_keys: str | Iterable[str] | None = None, overwrite: bool = False
    ) -> None:
        """Attach numeric values in :attr:`attrs` onto :attr:`data`.

        Iterate through values in :attr:`attrs` and attach `float` and `int` values
        to `data`.

        This method modifies object in place.

        Parameters
        ----------
        ignore_keys: str | Iterable[str], optional
            Do not broadcast selected keys.
            Defaults to None.
        overwrite : bool, optional
            If True, overwrite existing values in :attr:`data`. By default False.
        """
        if ignore_keys is None:
            ignore_keys = ()
        elif isinstance(ignore_keys, str):
            ignore_keys = (ignore_keys,)

        # Somewhat brittle: Only checking for int or float type
        numeric_attrs = [
            attr
            for attr, val in self.attrs.items()
            if (isinstance(val, (int, float)) and attr not in ignore_keys)
        ]
        self.broadcast_attrs(numeric_attrs, overwrite)

    # ------------
    # I / O
    # ------------

    def to_dataframe(self, copy: bool = True) -> pd.DataFrame:
        """Create DataFrame in which each key-value pair in :attr:`data` is a column.

        DataFrame does **not** copy data by default.
        Use the ``copy`` parameter to copy data values on creation.

        Parameters
        ----------
        copy : bool, optional
            Copy data on DataFrame creation.

        Returns
        -------
        pd.DataFrame
            DataFrame holding key-values as columns.
        """
        df = pd.DataFrame(self.data, copy=copy)
        df.attrs = self.attrs
        return df

    @classmethod
    def create_empty(
        cls: Type[VectorDatasetType],
        keys: Iterable[str],
        attrs: dict[str, Any] | None = None,
        **attrs_kwargs: Any,
    ) -> VectorDatasetType:
        """Create instance with variables defined by `keys` and size 0.

        If instance requires additional variables to be defined, these keys will automatically
        be attached to returned instance.

        Parameters
        ----------
        keys : Iterable[str]
            Keys to include in empty VectorDataset instance.
        attrs : dict[str, Any] | None, optional
            Attributes to attach instance.
        **attrs_kwargs : Any
            Define attributes as keyword arguments.

        Returns
        -------
        VectorDatasetType
            Empty VectorDataset instance.
        """
        return cls(data=_empty_vector_dict(keys or set()), attrs=attrs, copy=False, **attrs_kwargs)

    def generate_splits(
        self: VectorDatasetType, n_splits: int, copy: bool = True
    ) -> Generator[VectorDatasetType, None, None]:
        """Split instance into ``n_split`` sub-vectors.

        Parameters
        ----------
        n_splits : int
            Number of splits.
        copy : bool, optional
            Passed into :meth:`filter`. Defaults to True. Recommend to keep as True
            based on `numpy best practices <https://numpy.org/doc/stable/user/basics.indexing.html#slicing-and-striding>`_.

        Returns
        -------
        Generator[VectorDatasetType, None, None]
            Generator of split vectors.

        See Also
        --------
        :func:`numpy.array_split`
        """  # noqa: E501
        full_index = np.arange(self.size)
        index_splits = np.array_split(full_index, n_splits)
        for index in index_splits:
            filt = np.zeros(self.size, dtype=bool)
            filt[index] = True
            yield self.filter(filt, copy=copy)


class GeoVectorDataset(VectorDataset):
    """Base class to hold 1D geospatial arrays of consistent size.

    GeoVectorDataset is required to have geospatial coordinate keys defined
    in :attr:`required_keys`.

    Expect latitude-longitude CRS in WGS 84.
    Expect altitude in [:math:`m`].
    Expect level in [:math:`hPa`].

    Each spatial variable is expected to have "float32" or "float64" ``dtype``.
    The time variable is expected to have "datetime64[ns]" ``dtype``.

    Use the attribute :attr:`attr["crs"]` to specify coordinate reference system
    using `PROJ <https://proj.org/>`_ or `EPSG <https://epsg.org/home.html>`_ syntax.

    Parameters
    ----------
    data : dict[str, npt.ArrayLike] | pd.DataFrame | VectorDataDict | VectorDataset | None, optional
        Data dictionary or :class:`pandas.DataFrame` .
        Must include keys/columns ``time``, ``latitude``, ``longitude``, ``altitude`` or ``level``.
        Keyword arguments for ``time``, ``latitude``, ``longitude``, ``altitude`` or ``level``
        override ``data`` inputs. Expects ``altitude`` in meters and ``time``
        as a DatetimeLike (or array that can processed with :meth:`pd.to_datetime`).
        Additional waypoint-specific data can be included as additional keys/columns.
    longitude : npt.ArrayLike, optional
        Longitude data.
        Defaults to None.
    latitude : npt.ArrayLike, optional
        Latitude data.
        Defaults to None.
    altitude : npt.ArrayLike, optional
        Altitude data, [:math:`m`].
        Defaults to None.
    level : npt.ArrayLike, optional
        Level data, [:math:`hPa`].
        Defaults to None.
    time : npt.ArrayLike, optional
        Time data.
        Expects an array of DatetimeLike values,
        or array that can proccessed with :meth:`pd.to_datetime`.
        Defaults to None.
    attrs : dict[Hashable, Any] | AttrDict, optional
        Additional properties as a dictionary.
        Defaults to {}.
    copy : bool, optional
        Copy data on class creation.
        Defaults to True.
    **attrs_kwargs : Any
        Additional properties passed as keyword arguments.

    Raises
    ------
    KeyError
        Raises if ``data`` input does not contain at least ``time``, ``latitude``, ``longitude``,
        (``altitude`` or ``level``).
    """

    #: Required keys for creating GeoVectorDataset
    required_keys: set[str] = {
        "latitude",
        "longitude",
        "time",
    }

    def __init__(
        self,
        data: dict[str, npt.ArrayLike]
        | pd.DataFrame
        | VectorDataDict
        | VectorDataset
        | None = None,
        longitude: npt.ArrayLike | None = None,
        latitude: npt.ArrayLike | None = None,
        altitude: npt.ArrayLike | None = None,
        level: npt.ArrayLike | None = None,
        time: npt.ArrayLike | None = None,
        attrs: dict[str, Any] | AttrDict | None = None,
        copy: bool = True,
        **attrs_kwargs: Any,
    ) -> None:
        # shortcut to `GeoVectorDataset.create_empty` by just using `GeoVectorDataset()`
        if (
            data is None
            and longitude is None
            and latitude is None
            and altitude is None
            and level is None
            and time is None
        ):
            keys = self.required_keys.union(["altitude"])
            data = _empty_vector_dict(keys)

        super().__init__(data=data, attrs=attrs, copy=copy, **attrs_kwargs)

        # using the self[key] syntax specifically to run qc on assigment
        if longitude is not None:
            self["longitude"] = np.array(longitude, copy=copy)

        if latitude is not None:
            self["latitude"] = np.array(latitude, copy=copy)

        if time is not None:
            self["time"] = np.array(time, copy=copy)

        if altitude is not None:
            self["altitude"] = np.array(altitude, copy=copy)

        if level is not None:
            self["level"] = np.array(level, copy=copy)

        # warn if level and altitude are both defined
        if "level" not in self and "altitude" not in self and "altitude_ft" not in self:
            raise KeyError(
                f"{self.__class__.__name__} requires either "
                "`level` or `altitude` or `altitude_ft` data key"
            )

        # TODO: do we want to throw a warning if both level and altitude are specified?
        # This is the default behavior in Cocip
        # elif "level" in self and "altitude" in self:
        #     warnings.warn(
        #         f"{self.__class__.__name__} contains both `level` and `altitude` keys. "
        #         "This can result in ambiguous behavior."
        #     )

        # confirm that input dataframe has required columns
        if not self.required_keys.issubset(self):
            missing_keys = self.required_keys.difference(self)
            raise KeyError(
                f"Missing required data keys: {missing_keys}. "
                f"Use {self.__class__.__name__}.create_empty() to create an empty dataset."
            )

        # Parse time: If time is not np.datetime64, we try to coerce it to be
        # by pumping it through pd.to_datetime.
        time = self["time"]
        if not np.issubdtype(time.dtype, np.datetime64):
            warnings.warn("Time data is not np.datetime64. Attempting to coerce.")
            try:
                pd_time = pd.to_datetime(self["time"])
            except ValueError as e:
                raise ValueError("Could not coerce time data to datetime64.") from e
            np_time = pd_time.to_numpy(dtype="datetime64[ns]")
            self.update(time=np_time)
        elif time.dtype != "datetime64[ns]":
            self.update(time=time.astype("datetime64[ns]"))

        # Ensure spatial coordinates are float32 or float64
        float_dtype = (np.float32, np.float64)
        for coord in ("longitude", "latitude", "altitude", "level", "altitude_ft"):
            try:
                arr = self[coord]
            except KeyError:
                continue
            if arr.dtype not in float_dtype:
                self.update({coord: arr.astype(np.float64)})

        # set CRS to "EPSG:4326" by default
        if "crs" not in self.attrs:
            self.attrs["crs"] = "EPSG:4326"

        if self.attrs["crs"] == "EPSG:4326":
            if np.any(self["longitude"] > 180) or np.any(self["longitude"] < -180):
                raise ValueError("EPSG:4326 longitude coordinates should lie between [-180, 180).")
            if np.any(self["latitude"] > 90) or np.any(self["latitude"] < -90):
                raise ValueError("EPSG:4326 latitude coordinates should lie between [-90, 90].")

    @overrides
    def _display_attrs(self) -> dict[str, str]:
        try:
            time0, time1 = np.nanmin(self["time"]), np.nanmax(self["time"])
            lon0, lon1 = np.nanmin(self["longitude"]), np.nanmax(self["longitude"])
            lat0, lat1 = np.nanmin(self["latitude"]), np.nanmax(self["latitude"])
            alt0, alt1 = np.nanmin(self.altitude), np.nanmax(self.altitude)
            attrs = {
                "time": f"[{pd.Timestamp(time0)}, {pd.Timestamp(time1)}]",
                "longitude": f"[{lon0}, {lon1}]",
                "latitude": f"[{lat0}, {lat1}]",
                "altitude": f"[{alt0}, {alt1}]",
            }
        except Exception:
            attrs = {}

        attrs.update(super()._display_attrs())
        return attrs

    @property
    def level(self) -> npt.NDArray[np.float_]:
        """Get pressure ``level`` values for points.

        Automatically calculates pressure level using :func:`units.m_to_pl` using ``altitude`` key.

        Note that if ``level`` key exists in :attr:`data`, the data at the ``level``
        key will be returned. This allows an override of the default calculation
        of pressure level from altitude.

        Returns
        -------
        npt.NDArray[np.float_]
            Point pressure level values, [:math:`hPa`]
        """
        try:
            return self["level"]
        except KeyError:
            return units.m_to_pl(self.altitude)

    @property
    def altitude(self) -> npt.NDArray[np.float_]:
        """Get altitude.

        Automatically calculates altitude using :func:`units.pl_to_m` using ``level`` key.

        Note that if ``altitude`` key exists in :attr:`data`, the data at the ``altitude``
        key will be returned. This allows an override of the default calculation of altitude
        from pressure level.

        Returns
        -------
        npt.NDArray[np.float_]
            Altitude, [:math:`m`]
        """
        try:
            return self["altitude"]
        except KeyError:
            # Implementation note: explicitly look for "level" or "altitude_ft" key
            # here to avoid getting stuck in an infinite loop when .level or .altitude_ft
            # are called.
            if (level := self.get("level")) is not None:
                return units.pl_to_m(level)
            return units.ft_to_m(self["altitude_ft"])

    @property
    def air_pressure(self) -> npt.NDArray[np.float_]:
        """Get ``air_pressure`` values for points.

        Returns
        -------
        npt.NDArray[np.float_]
            Point air pressure values, [:math:`Pa`]
        """
        try:
            return self["air_pressure"]
        except KeyError:
            return 100 * self.level

    @property
    def altitude_ft(self) -> npt.NDArray[np.float_]:
        """Get altitude in feet.

        Returns
        -------
        npt.NDArray[np.float_]
            Altitude, [:math:`ft`]
        """
        try:
            return self["altitude_ft"]
        except KeyError:
            return units.m_to_ft(self.altitude)

    @property
    def constants(self) -> dict[str, Any]:
        """Return a dictionary of constant attributes and data values.

        Includes :attr:`attrs` and values from columns in :attr:`data` with a unique
        value.

        Returns
        -------
        dict[str, Any]
            Properties and their constant values
        """
        constants = {}

        # get constant data values that are not nan
        for key in set(self.data) - self.required_keys:
            unique = np.unique(self[key])
            if len(unique) == 1 and (isinstance(unique[0], str) or ~np.isnan(unique[0])):
                constants[key] = unique[0]

        # add attributes
        constants.update(self.attrs)

        # clean strings values by removing whitespace
        # convert any numpy items to python objects
        def _cleanup(v: Any) -> Any:
            if isinstance(v, str):
                return v.strip()
            if isinstance(v, np.integer):
                return int(v)
            if isinstance(v, np.floating):
                return float(v)
            if isinstance(v, np.bool_):
                return bool(v)
            return v

        return {k: _cleanup(v) for k, v in constants.items()}

    @property
    def coords(self) -> dict[str, np.ndarray]:
        """Get geospatial coordinates for compatibility with MetDataArray.

        Returns
        -------
        pd.DataFrame
            :class:`pd.DataFrame` with columns `longitude`, `latitude`, `level`, and `time`.
        """
        return {
            "longitude": self["longitude"],
            "latitude": self["latitude"],
            "level": self.level,
            "time": self["time"],
        }

    # ------------
    # Utilities
    # ------------

    def transform_crs(
        self: GeoVectorDatasetType, crs: str, copy: bool = True
    ) -> GeoVectorDatasetType:
        """Transform trajectory data from one coordinate reference system (CRS) to another.

        Parameters
        ----------
        crs : str
            Target CRS. Passed into to :class:`pyproj.Transformer`. The source CRS
            is inferred from the :attr:`attrs["crs"]` attribute.
        copy : bool, optional
            Copy data on transformation. Defaults to True.

        Returns
        -------
        GeoVectorDatasetType
            Converted dataset with new coordinate reference system.
            :attr:`attrs["crs"]` reflects new crs.
        """
        transformer = pyproj.Transformer.from_crs(self.attrs["crs"], crs, always_xy=True)
        lon, lat = transformer.transform(self["longitude"], self["latitude"])

        if copy:
            ret = self.copy()
        else:
            ret = self

        ret.update(longitude=lon, latitude=lat)
        ret.attrs.update(crs=crs)
        return ret

    # ------------
    # Met
    # ------------

    def coords_intersect_met(
        self, met: met_module.MetDataset | met_module.MetDataArray
    ) -> npt.NDArray[np.bool_]:
        """Return boolean mask of data inside the bounding box defined by ``met``.

        Parameters
        ----------
        met : MetDataset | MetDataArray
            MetDataset or MetDataArray to compare.

        Returns
        -------
        npt.NDArray[np.bool_]
            True if point is inside the bounding box defined by ``met``.
        """

        lat_intersect = coordinates.intersect_domain(
            met.variables["latitude"].values,
            self["latitude"],
        )
        lon_intersect = coordinates.intersect_domain(
            met.variables["longitude"].values,
            self["longitude"],
        )
        level_intersect = coordinates.intersect_domain(
            met.variables["level"].values,
            self.level,
        )
        time_intersect = coordinates.intersect_domain(
            met.variables["time"].values,
            self["time"],
        )

        return lat_intersect & lon_intersect & level_intersect & time_intersect

    def intersect_met(
        self,
        mda: met_module.MetDataArray,
        *,
        longitude: npt.NDArray[np.float_] | None = None,
        latitude: npt.NDArray[np.float_] | None = None,
        level: npt.NDArray[np.float_] | None = None,
        time: npt.NDArray[np.datetime64] | None = None,
        use_indices: bool = False,
        **interp_kwargs: Any,
    ) -> npt.NDArray[np.float_]:
        """Intersect waypoints with MetDataArray.

        Parameters
        ----------
        mda : MetDataArray
            MetDataArray containing a meteorological variable at spatio-temporal coordinates.
        longitude : npt.NDArray[np.float_], optional
            Override existing coordinates for met interpolation
        latitude : npt.NDArray[np.float_], optional
            Override existing coordinates for met interpolation
        level : npt.NDArray[np.float_], optional
            Override existing coordinates for met interpolation
        time : npt.NDArray[np.datetime64], optional
            Override existing coordinates for met interpolation
        use_indices : bool, optional
            Experimental.
        **interp_kwargs : Any
            Additional keyword arguments to pass to :meth:`MetDataArray.intersect_met`.
            Examples include ``method``, ``bounds_error``, and ``fill_value``. If an error such as
            `ValueError: One of the requested xi is out of bounds in dimension 2` occurs, try
            calling this function with ``bounds_error=False``. In addition, setting
            ``fill_value=0.0`` will replace NaN values with 0.0.

        Returns
        -------
        npt.NDArray[np.float_]
            Interpolated values

        Examples
        --------
        >>> from datetime import datetime
        >>> import pandas as pd
        >>> import numpy as np
        >>> from pycontrails.datalib.ecmwf import ERA5
        >>> from pycontrails import Flight

        >>> # Get met data
        >>> times = (datetime(2022, 3, 1, 0),  datetime(2022, 3, 1, 3))
        >>> variables = ["air_temperature", "specific_humidity"]
        >>> levels = [300, 250, 200]
        >>> era5 = ERA5(time=times, variables=variables, pressure_levels=levels)
        >>> met = era5.open_metdataset()

        >>> # Example flight
        >>> df = pd.DataFrame()
        >>> df['longitude'] = np.linspace(0, 50, 10)
        >>> df['latitude'] = np.linspace(0, 10, 10)
        >>> df['altitude'] = 11000
        >>> df['time'] = pd.date_range("2022-03-01T00", "2022-03-01T02", periods=10)
        >>> fl = Flight(df)

        >>> # Intersect
        >>> fl.intersect_met(met['air_temperature'], method='nearest')
        array([231.6297 , 230.72604, 232.2432 , 231.88339, 231.0643 , 231.59073,
               231.65126, 231.93065, 232.03345, 231.65955], dtype=float32)

        >>> fl.intersect_met(met['air_temperature'], method='linear')
        array([225.77794, 225.13908, 226.23122, 226.31831, 225.56102, 225.81192,
               226.03194, 226.22057, 226.0377 , 225.63226], dtype=float32)

        >>> # Interpolate and attach to `Flight` instance
        >>> for key in met:
        ...     fl[key] = fl.intersect_met(met[key])

        >>> # Show the final three columns of the dataframe
        >>> fl.dataframe.iloc[:, -3:].head()
                         time  air_temperature  specific_humidity
        0 2022-03-01 00:00:00       225.777939           0.000132
        1 2022-03-01 00:13:20       225.139084           0.000132
        2 2022-03-01 00:26:40       226.231216           0.000107
        3 2022-03-01 00:40:00       226.318314           0.000171
        4 2022-03-01 00:53:20       225.561020           0.000109

        """
        # Override use_indices in certain situations
        if use_indices:
            # Often the single_level data we use has time shifted
            # Don't allow it for now. We could do something smarter here!
            if mda.is_single_level:
                use_indices = False

            # Cannot both override some coordinate AND pass indices.
            elif any(c is not None for c in (longitude, latitude, level, time)):
                # Should we warn?! Or is this "convenience"?
                use_indices = False

        longitude = longitude if longitude is not None else self["longitude"]
        latitude = latitude if latitude is not None else self["latitude"]
        level = level if level is not None else self.level
        time = time if time is not None else self["time"]

        if not use_indices:
            return mda.interpolate(longitude, latitude, level, time, **interp_kwargs)

        indices = self._get_indices()
        already_has_indices = indices is not None
        out, indices = mda.interpolate(
            longitude,
            latitude,
            level,
            time,
            indices=indices,
            return_indices=True,
            **interp_kwargs,
        )
        if not already_has_indices:
            self._put_indices(indices)
        return out

    def _put_indices(self, indices: interpolation.RGIArtifacts) -> None:
        """Set entries of ``indices`` onto underlying :attr:`data.

        Each entry of ``indices`` are unpacked assuming certain conventions
        for its structure. A ValueError is raise if these conventions are not
        satisfied.

        .. versionadded:: 0.26.0

            Experimental


        Parameters
        ----------
        indices : interpolation.RGIArtifacts
            The indices to store.
        """
        indices_x, indices_y, indices_z, indices_t = indices.xi_indices
        distances_x, distances_y, distances_z, distances_t = indices.norm_distances
        out_of_bounds = indices.out_of_bounds

        self["_indices_x"] = indices_x
        self["_indices_y"] = indices_y
        self["_indices_z"] = indices_z
        self["_indices_t"] = indices_t
        self["_distances_x"] = distances_x
        self["_distances_y"] = distances_y
        self["_distances_z"] = distances_z
        self["_distances_t"] = distances_t
        self["_out_of_bounds"] = out_of_bounds

    def _get_indices(self) -> interpolation.RGIArtifacts | None:
        """Get entries from call to :meth:`_put_indices`.

        .. versionadded:: 0.26.0

            Experimental

        Returns
        -------
        tuple | None
            Previously cached output of
            :meth:`scipy.interpolate.RegularGridInterpolator._find_indices`,
            or None if cached output is not present on instance.
        """
        try:
            indices_x = self["_indices_x"]
            indices_y = self["_indices_y"]
            indices_z = self["_indices_z"]
            indices_t = self["_indices_t"]
            distances_x = self["_distances_x"]
            distances_y = self["_distances_y"]
            distances_z = self["_distances_z"]
            distances_t = self["_distances_t"]
            out_of_bounds = self["_out_of_bounds"]
        except KeyError:
            return None

        indices = np.asarray([indices_x, indices_y, indices_z, indices_t])
        distances = np.asarray([distances_x, distances_y, distances_z, distances_t])

        return interpolation.RGIArtifacts(indices, distances, out_of_bounds)

    def _invalidate_indices(self) -> None:
        """Remove any cached indices from :attr:`data."""
        for key in (
            "_indices_x",
            "_indices_y",
            "_indices_z",
            "_indices_t",
            "_distances_x",
            "_distances_y",
            "_distances_z",
            "_distances_t",
            "_out_of_bounds",
        ):
            self.data.pop(key, None)

    @overload
    def downselect_met(
        self,
        met: met_module.MetDataset,
        *,
        longitude_buffer: tuple[float, float] = ...,
        latitude_buffer: tuple[float, float] = ...,
        level_buffer: tuple[float, float] = ...,
        time_buffer: tuple[np.timedelta64, np.timedelta64] = ...,
        copy: bool = ...,
    ) -> met_module.MetDataset:
        ...

    @overload
    def downselect_met(
        self,
        met: met_module.MetDataArray,
        *,
        longitude_buffer: tuple[float, float] = ...,
        latitude_buffer: tuple[float, float] = ...,
        level_buffer: tuple[float, float] = ...,
        time_buffer: tuple[np.timedelta64, np.timedelta64] = ...,
        copy: bool = ...,
    ) -> met_module.MetDataArray:
        ...

    def downselect_met(
        self,
        met: met_module.MetDataType,
        *,
        longitude_buffer: tuple[float, float] = (0.0, 0.0),
        latitude_buffer: tuple[float, float] = (0.0, 0.0),
        level_buffer: tuple[float, float] = (0.0, 0.0),
        time_buffer: tuple[np.timedelta64, np.timedelta64] = (
            np.timedelta64(0, "h"),
            np.timedelta64(0, "h"),
        ),
        copy: bool = True,
    ) -> met_module.MetDataType:
        """Downselect ``met`` to encompass a spatiotemporal region of the data.

        Parameters
        ----------
        met : MetDataset | MetDataArray
            MetDataset or MetDataArray to downselect.
        longitude_buffer : tuple[float, float], optional
            Extend longitude domain past by ``longitude_buffer[0]`` on the low side
            and ``longitude_buffer[1]`` on the high side.
            Units must be the same as class coordinates.
            Defaults to ``(0, 0)`` degrees.
        latitude_buffer : tuple[float, float], optional
            Extend latitude domain past by ``latitude_buffer[0]`` on the low side
            and ``latitude_buffer[1]`` on the high side.
            Units must be the same as class coordinates.
            Defaults to ``(0, 0)`` degrees.
        level_buffer : tuple[float, float], optional
            Extend level domain past by ``level_buffer[0]`` on the low side
            and ``level_buffer[1]`` on the high side.
            Units must be the same as class coordinates.
            Defaults to ``(0, 0)`` [:math:`hPa`].
        time_buffer : tuple[np.timedelta64, np.timedelta64], optional
            Extend time domain past by ``time_buffer[0]`` on the low side
            and ``time_buffer[1]`` on the high side.
            Units must be the same as class coordinates.
            Defaults to ``(np.timedelta64(0, "h"), np.timedelta64(0, "h"))``.
        copy : bool
            If returned object is a copy or view of the original. True by default.

        Returns
        -------
        MetDataset | MetDataArray
            Copy of downselected MetDataset or MetDataArray.
        """
        lon_slice = coordinates.slice_domain(
            met.variables["longitude"].values,
            self["longitude"],
            buffer=longitude_buffer,
        )
        lat_slice = coordinates.slice_domain(
            met.variables["latitude"].values,
            self["latitude"],
            buffer=latitude_buffer,
        )
        time_slice = coordinates.slice_domain(
            met.variables["time"].values,
            self["time"],
            buffer=time_buffer,
        )

        # single level data have "level" == [-1]
        if met.is_single_level:
            level_slice = slice(None)
        else:
            level_slice = coordinates.slice_domain(
                met.variables["level"].values,
                self.level,
                buffer=level_buffer,
            )
        logger.debug("Downselect met at %s %s %s %s", lon_slice, lat_slice, level_slice, time_slice)

        data = met.data.isel(
            longitude=lon_slice,
            latitude=lat_slice,
            level=level_slice,
            time=time_slice,
        )
        return type(met)(data, copy=copy)

    # ------------
    # I / O
    # ------------

    @classmethod
    @overrides
    def create_empty(
        cls: Type[GeoVectorDatasetType],
        keys: Iterable[str] | None = None,
        attrs: dict[str, Any] | None = None,
        **attrs_kwargs: Any,
    ) -> GeoVectorDatasetType:
        keys = cls.required_keys.union(keys or set())
        keys.add("altitude")
        return super().create_empty(keys, attrs, **attrs_kwargs)

    def to_geojson_points(self) -> dict[str, Any]:
        """Return dataset as GeoJSON FeatureCollection of Points.

        Each Feature has a properties attribute that includes `time` and
        other data besides `latitude`, `longitude`, and `altitude` in :attr:`data`.

        Returns
        -------
        dict[str, Any]
            Python representation of GeoJSON FeatureCollection
        """
        return json_module.dataframe_to_geojson_points(self.dataframe)

    def to_pseudo_mercator(self: GeoVectorDatasetType, copy: bool = True) -> GeoVectorDatasetType:
        """Convert data from :attr:`attrs["crs"]` to Pseudo Mercator (EPSG:3857).

        Parameters
        ----------
        copy : bool, optional
            Copy data on tranformation.
            Defaults to True.

        Returns
        -------
        GeoVectorDatasetType
        """
        return self.transform_crs("EPSG:3857", copy=copy)



"""A single data structure encompassing a sequence of :class:`Flight` instances."""

from __future__ import annotations

from typing import Any, Iterable

import numpy as np
import numpy.typing as npt
import pandas as pd
from overrides import overrides

from pycontrails.core.flight import Flight
from pycontrails.core.fuel import Fuel, JetA
from pycontrails.core.vector import GeoVectorDataset


class Fleet(Flight):
    """Data structure for holding a sequence of `Flight` instances.

    Flight waypoints are merged into a single `Flight`-like object.
    """

    def __init__(
        self,
        data: dict[str, npt.ArrayLike] | None = None,
        longitude: npt.ArrayLike | None = None,
        latitude: npt.ArrayLike | None = None,
        altitude: npt.ArrayLike | None = None,
        level: npt.ArrayLike | None = None,
        time: npt.ArrayLike | None = None,
        attrs: dict[str, Any] | None = None,
        copy: bool = True,
        fuel: Fuel | None = None,
        **attrs_kwargs: Any,
    ) -> None:
        # Do not want to call Flight.__init__
        # The Flight constructor assumes a sorted time column
        GeoVectorDataset.__init__(
            self,
            data=data,
            longitude=longitude,
            latitude=latitude,
            altitude=altitude,
            level=level,
            time=time,
            attrs=attrs,
            copy=copy,
            **attrs_kwargs,
        )

        if "fl_attrs" not in self.attrs:
            raise ValueError("Require key 'fl_attrs' in Fleet attrs.")

        # set fuel for ALL FLIGHTS
        # NOTE: We could also set fuel for each flight via the `fl_attrs` attribute
        if fuel is None:
            self.fuel = JetA()
        else:
            self.fuel = fuel

        self.final_waypoints = self.calc_final_waypoints()

    def calc_final_waypoints(self) -> npt.NDArray[np.bool_]:
        """Validate data and calculate the final waypoint of each flight.

        Returns
        -------
        npt.NDArray[np.bool_]
            A boolean array in which True values correspond to final waypoint of each flight.

        Raises
        ------
        KeyError, ValueError
            Fleet :attr:`data` does not take the expected form.
        """
        try:
            flight_id = self["flight_id"]
        except KeyError:
            raise KeyError("Expect `flight_id` in Fleet data.") from None

        # Some pandas groupby magic to ensure flights are arranged in blocks
        df = pd.DataFrame({"flight_id": flight_id}).reset_index()
        grouped = df.groupby("flight_id", sort=False)
        groups = grouped.agg({"flight_id": "size", "index": ["first", "last"]})
        expected_size = groups[("index", "last")] - groups[("index", "first")] + 1
        actual_size = groups[("flight_id", "size")]
        if not (expected_size == actual_size).all():
            raise ValueError("Fleet must have contiguous waypoints blocks with constant flight_id")

        # Return boolean array of final waypoints by flight
        final_waypoints = np.zeros(self.size, dtype=bool)
        final_waypoint_indices = groups[("index", "last")].to_numpy()
        final_waypoints[final_waypoint_indices] = True
        return final_waypoints

    def fit_altitude(
        self,
        max_segments: int = 30,
        pop: int = 3,
        r2_target: float = 0.999,
        max_cruise_rocd: float = 10,
        sg_window: int = 7,
        sg_polyorder: int = 1,
    ) -> Fleet:
        """Use piecewise linear fitting to smooth a flight profile.

        Fit a flight profile to a series of line segments. Segments that have a
        small rocd will be set to have a slope of zero and snapped to the
        nearest thousand foot level.  A Savitzky-Golay filter will then be
        applied to the profile to smooth the climbs and descents.  This filter
        works best for high frequency flight data, sampled at a 1-3 second
        sampling period.

        Parameters
        ----------
        max_segments : int, optional
            The maximum number of line segements to fit to the flight profile.
        pop: int, optional
            Population parameter used for the stocastic optimization routine
            used to fit the flight profile.
        r2_target: float, optional
            Target r^2 value for solver. Solver will continue to add line
            segments until the resulting r^2 value is greater than this.
        max_cruise_rocd: float, optional
            The maximum ROCD for a segment that will be forced to a slope of
            zero, [:math:`ft s^{-1}`]
        sg_window: int, optional
            Parameter for :func:`scipy.signal.savgol_filter`
        sg_polyorder: int, optional
            Parameter for :func:`scipy.signal.savgol_filter`

        Returns
        -------
        Fleet
            Smoothed flight
        """
        raise NotImplementedError("Only implemented for Flight instances")

    @classmethod
    def from_seq(
        cls,
        seq: Iterable[Flight],
        broadcast_numeric: bool = True,
        copy: bool = True,
        attrs: dict[str, Any] | None = None,
    ) -> Fleet:
        """Instantiate a `Fleet` instance from an iterable of Flight.

        The entire sequence ``seq`` is cast to a list, and so it's preferable for
        the parameter ``seq`` to come in as a list.

        Parameters
        ----------
        seq : Iterable[Flight]
            An iterable of `Flight` instances.
        broadcast_numeric : bool, optional
            If True, broadcast numeric attributes to data variables.
        copy : bool, optional
            If True, make copy of each flight instance in ``seq``.
        attrs : dict[str, Any] | None, optional
            Global attribute to attach to instance. Must NOT contain the "fl_attrs"
            key (this is automatically deduced when iterating over ``seq``).

        Returns
        -------
        Fleet
            A `Fleet` instance made from concatenating the :class:`Flight`
            instances in ``seq``.
        """
        if attrs is None:
            attrs = {}
        if "fl_attrs" in attrs:
            raise ValueError("Parameter 'attrs' cannot contain 'fl_attrs' key.")
        attrs["fl_attrs"] = {}
        attrs.setdefault("data_keys", None)
        attrs.setdefault("crs", None)

        if copy:
            seq = [fl.copy() for fl in seq]
        else:
            seq = list(seq)

        for fl in seq:
            cls._validate_fl(fl, seq[0].fuel, attrs, broadcast_numeric)

        # Surprisingly, it's faster to call np.concatenate on each variable
        # separately then it is to call pd.concat on the sequence of fl.dataframe.
        # We do this concatenation below.
        data = {var: np.concatenate([fl[var] for fl in seq]) for var in attrs["data_keys"]}
        return cls(data=data, attrs=attrs, copy=False)

    @staticmethod
    def _validate_fl(
        fl: Flight, fuel: Fuel, attrs: dict[str, Any], broadcast_numeric: bool
    ) -> None:
        """Attach "flight_id" and "waypoint" columns to flight :attr:`data`.

        Mutates parameters `fl` and `attrs` in place.

        Parameters
        ----------
        fl : Flight
            Flight instance to process.
        fuel : Fuel
            Fuel used all flights
        attrs : dict[str, Any]
            Dictionary of `Fleet` attributes. Attributes belonging to `fl` are attached
            to `attrs`.
        broadcast_numeric : bool
            If True, broadcast numeric attributes to data variables.

        Raises
        ------
        KeyError
            `fl` does not have a `flight_id` in :attr:`attrs`.
        ValueError
            If `flight_id` is duplicated or incompatible CRS found.
        AttributeError
            If `fuel` is incompatible with other flights.
        """
        # Validate and cache attrs
        if "flight_id" not in fl.attrs:
            raise KeyError("Each flight must have `flight_id` in its `attrs`.")
        flight_id = fl.attrs["flight_id"]

        if flight_id in attrs["fl_attrs"]:
            raise ValueError(f"Duplicate `flight_id` {flight_id} found.")
        attrs["fl_attrs"][flight_id] = fl.attrs

        # Verify fuel type is consistent across flights
        if fl.fuel != fuel:
            raise AttributeError(
                f"Fuel type on Flight {flight_id} ({fl.fuel.fuel_name}) "
                f"is not inconsistent with previous flights ({fuel.fuel_name}). "
                "The `fuel` attributes must be consistent between flights in a Fleet."
            )

        # Verify CRS
        crs = fl.attrs["crs"]
        if attrs["crs"] is None:
            attrs["crs"] = crs
        elif attrs["crs"] != crs:
            raise ValueError(f"Incompatible CRS {attrs['crs']} and {crs} found among flights.")

        # Expand data
        if broadcast_numeric:
            fl.broadcast_numeric_attrs(ignore_keys=["load_factor"])
        if "waypoint" not in fl:
            fl["waypoint"] = np.arange(fl.size)
        if "flight_id" not in fl:
            fl["flight_id"] = np.full(fl.size, flight_id)

        # Verify consistent data keys
        # When attaching data_keys to attrs, use set(data_keys)
        # But, we don't need to create this set separately for each flight
        # Simply use fl.data.keys() to check for consistency.
        data_keys = fl.data.keys()
        if attrs["data_keys"] is None:
            attrs["data_keys"] = set(data_keys)
        elif attrs["data_keys"] != data_keys:
            raise ValueError(
                f"Inconsistent data keys {attrs['data_keys']}  and {data_keys} found among flights."
            )

    @property
    def n_flights(self) -> int:
        """Return number of distinct flights.

        Returns
        -------
        int
            Number of flights
        """
        return len(self.attrs["fl_attrs"])

    def to_flight_list(self, copy: bool = True) -> list[Flight]:
        """De-concatenate merged waypoints into a list of Flight instances.

        Any global :attr:`attrs` are lost.

        Parameters
        ----------
        copy : bool, optional
            If True, make copy of each flight instance in `seq`.

        Returns
        -------
        list[Flight]
            List of Flights in the same order as was passed into the `Fleet` instance.
        """
        grouped = self.dataframe.groupby("flight_id", sort=False)
        return [
            Flight(df, attrs=self.attrs["fl_attrs"][flight_id], copy=copy)
            for flight_id, df in grouped
        ]

    ###################################
    # Flight methods involving segments
    ###################################

    def segment_true_airspeed(
        self,
        u_wind: npt.NDArray[np.float_] | float = 0.0,
        v_wind: npt.NDArray[np.float_] | float = 0.0,
        smooth: bool = True,
        window_length: int = 7,
        polyorder: int = 1,
    ) -> npt.NDArray[np.float_]:
        """Calculate the true airspeed [:math:`m / s`] from the ground speed and horizontal winds.

        Because Flight.segment_true_airspeed uses a smoothing pattern, waypoints in :attr:`data`
        are not independent. Moreover, we expect the final waypoint of each flight to have a nan
        value associated to any segment property. Consequently, we need to define a custom method
        here to deal with these issues when applying this method on a fleet of flights.

        See docstring for :meth:`Flight.segment_true_airspeed`.

        Raises
        ------
        RuntimeError
            Unexpected key `__u_wind` or `__v_wind` found in :attr:`data`.
        """
        if isinstance(u_wind, np.ndarray):
            # Choosing a key we don't think exists
            key = "__u_wind"
            if key in self:
                raise RuntimeError(f"Unexpected key {key} found")
            self[key] = u_wind

        if isinstance(v_wind, np.ndarray):
            # Choosing a key we don't think exists
            key = "__v_wind"
            if key in self:
                raise RuntimeError(f"Unexpected key {key} found")
            self[key] = v_wind

        # Calculate TAS on each flight individually
        def calc_tas(fl: Flight) -> npt.NDArray[np.float_]:
            u = fl.get("__u_wind", u_wind)
            v = fl.get("__v_wind", v_wind)

            return fl.segment_true_airspeed(
                u, v, smooth=smooth, window_length=window_length, polyorder=polyorder
            )

        fls = self.to_flight_list(copy=False)
        tas = [calc_tas(fl) for fl in fls]

        # Cleanup
        self.data.pop("__u_wind", None)
        self.data.pop("__v_wind", None)

        # Making an assumption here that Fleet was instantiated by `from_seq`
        # method. If this is not the case, the order may be off when to_flight_list
        # is called.
        # Currently, we expect to only use Fleet "internally", so this more general
        # use case isn't seen.
        return np.concatenate(tas)

    @overrides
    def segment_groundspeed(self, *args: Any, **kwargs: Any) -> npt.NDArray[np.float_]:
        # Implement if we have a usecase for this.
        # Because the super() method uses a smoothing pattern, it will not reliably
        # work on Fleet.
        raise NotImplementedError

    @overrides
    def resample_and_fill(self, *args: Any, **kwargs: Any) -> Fleet:
        # Definitely do not try to call this on a Fleet! Lots of trajectory
        # specific logic is employed on the Fligth method.
        raise NotImplementedError

    @overrides
    def segment_length(self) -> npt.NDArray[np.float_]:
        return np.where(self.final_waypoints, np.nan, super().segment_length())

    @property
    @overrides
    def max_distance_gap(self) -> float:
        if self.attrs["crs"] != "EPSG:4326":
            raise NotImplementedError("Only implemented for EPSG:4326 CRS.")

        return np.nanmax(self.segment_length()).item()

    @overrides
    def segment_azimuth(self) -> npt.NDArray[np.float_]:
        return np.where(self.final_waypoints, np.nan, super().segment_azimuth())

    @overrides
    def segment_angle(self) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.float_]]:
        sin_a, cos_a = super().segment_angle()
        sin_a[self.final_waypoints] = np.nan
        cos_a[self.final_waypoints] = np.nan
        return sin_a, cos_a



"""Emissions model tests."""

from __future__ import annotations

import numpy as np
import pytest

from pycontrails import Flight
from pycontrails.core.fuel import JetA
from pycontrails.core.models import Model
from pycontrails.models.emissions import Emissions
from pycontrails.models.emissions import black_carbon as nvpm
from pycontrails.models.emissions import emissions as emissions_mod
from pycontrails.models.emissions import ffm2
from pycontrails.physics import jet, units
from pycontrails.physics.jet import thrust_setting_nd


def test_emissions_init():
    """Test emissions __init__ method."""
    emissions = Emissions()
    assert "01P11CM114" in emissions.edb_engine_gaseous  # Engine for Boeing 737-700 (CFM56-7B24E)
    assert "01P11CM114" in emissions.edb_engine_nvpm  # Engine for Boeing 737-700 (CFM56-7B24E)

    assert "3GE074" in emissions.edb_engine_gaseous  # Gaseous data for old engine (DC10, CF6-50C2)
    assert "3GE074" not in emissions.edb_engine_nvpm  # No nvPM data for old engine (DC10, CF6-50C2)
    assert not hasattr(emissions, "source")

    assert isinstance(emissions, Model)
    assert isinstance(emissions.params, dict)


def test_emissions_class_variables():
    """Check that `Emissions` class only creates dictionaries once."""
    emissions1 = Emissions()
    emissions2 = Emissions()
    assert emissions1 is not emissions2
    assert emissions1.edb_engine_gaseous is emissions2.edb_engine_gaseous
    assert emissions2.edb_engine_nvpm is emissions2.edb_engine_nvpm

    # Update if emissions data changes
    assert len(emissions1.edb_engine_gaseous) == 557
    assert len(emissions1.edb_engine_nvpm) == 178

    # Check that the gaseous engine UIDs are a superset of the nvPM engine UIDs in the EDB.
    # This logic is used in cocip_grid.calc_emissions
    assert set(emissions1.edb_engine_gaseous).issuperset(emissions1.edb_engine_nvpm)


@pytest.mark.parametrize("aircraft_type", ["B737", "B738", "A320"])
@pytest.mark.parametrize("engine_uid", ["01P11CM114", "01P11CM116", "01P08CM105"])
@pytest.mark.parametrize("bada", ["BADA3", "BADA4"])
def test_emissions_eval(flight_fake: Flight, bada: str, aircraft_type: str, engine_uid: str):
    """Test emissions `eval` method.

    Simple smoke test with several different BADA sources and aircraft types.
    """
    emissions = Emissions()

    match = "Variable `air_temperature` not found. "
    with pytest.raises(KeyError, match=match):
        emissions.eval(source=flight_fake)

    flight_fake["air_temperature"] = 216 * np.ones(flight_fake.size)
    flight_fake["specific_humidity"] = 1e-6 * np.ones(flight_fake.size)
    flight_fake["true_airspeed"] = 247 * np.ones(flight_fake.size)
    flight_fake["fuel_flow"] = 0.3 * np.ones(flight_fake.size)
    flight_fake.attrs["aircraft_type"] = aircraft_type
    flight_fake.attrs["aircraft_type_bada"] = aircraft_type
    flight_fake.attrs["engine_uid"] = engine_uid
    flight_fake.attrs["bada_model"] = bada
    flight_fake.attrs["n_engine"] = 2

    out_fl = emissions.eval(source=flight_fake)
    assert emissions.source is not flight_fake  # since copy is True
    assert isinstance(out_fl, Flight)

    # Constants in => nvpm_ei_n is now a function of altitude
    counts = np.unique(out_fl["nvpm_ei_n"], return_counts=True)[1]
    assert counts.size == 500

    # ensure that "thrust", "nvpm_ei_n" and "nvpm_ei_m" are not overwritten
    flight_fake["thrust_setting"] = 0.3 * np.ones(flight_fake.size)
    out_fl = emissions.eval(source=flight_fake)
    assert np.all(out_fl["thrust_setting"] == 0.3)
    del flight_fake["thrust_setting"]

    # both nvpm_ei_n keys need to be defined here
    flight_fake["nvpm_ei_n"] = 1e15 * np.ones(flight_fake.size)
    out_fl = emissions.eval(source=flight_fake)

    assert np.all(out_fl["nvpm_ei_n"] == 1e15)

    flight_fake["nvpm_ei_m"] = 2e-4 * np.ones(flight_fake.size)
    out_fl = emissions.eval(flight_fake)

    assert np.all(out_fl["nvpm_ei_n"] == 1e15)
    assert np.all(out_fl["nvpm_ei_m"] == 2e-4)

    # TODO: We don't currently guard against overwriting on the pollutants
    flight_fake["co2"] = 15 * np.ones(flight_fake.size)
    with pytest.warns(UserWarning, match="Overwriting data in key `co2`"):
        out_fl = emissions.eval(flight_fake)


def test_emissions_index_ffm2():
    """Test emissions without explicit flight.

    Pin a few values that fit expectations.
    """
    engine_uid = "01P18RR103"  # Engine for Airbus A380 (Trent 970-84)

    # Cruise conditions
    air_pressure = 19650  # Units: Pa
    air_temperature = 216  # Units: K
    true_airspeed = 247.5  # Units: m/s
    fuel_flow_per_engine_cruise = 0.882  # Units: kg/s
    fuel_flow_per_engine_descent = 0.125  # Units: kg/s (Set to < 7% equivalent engine power)

    emissions = Emissions()
    assert not hasattr(emissions, "source")
    edb_gaseous = emissions.edb_engine_gaseous[engine_uid]

    # Nitrogen oxide
    nox_ei = emissions_mod.nitrogen_oxide_emissions_index_ffm2(
        edb_gaseous,
        fuel_flow_per_engine_cruise,
        true_airspeed,
        air_pressure,
        air_temperature,
    )
    np.testing.assert_almost_equal(nox_ei, np.array([0.019298]), decimal=6)

    # Carbon monoxide (descent conditions with very low power)
    co_ei = emissions_mod.carbon_monoxide_emissions_index_ffm2(
        edb_gaseous,
        fuel_flow_per_engine_descent,
        true_airspeed,
        air_pressure,
        air_temperature,
    )
    np.testing.assert_almost_equal(co_ei, np.array([0.028461]), decimal=6)

    # Hydrocarbons (descent conditions with very low power)
    hc_ei = emissions_mod.hydrocarbon_emissions_index_ffm2(
        edb_gaseous,
        fuel_flow_per_engine_descent,
        true_airspeed,
        air_pressure,
        air_temperature,
    )
    np.testing.assert_almost_equal(hc_ei, np.array([0.000077817]), decimal=6)


def test_E45x_hydrocarbon_emissions_index() -> None:
    """Test specific error arising from duplicate indexes in co_hc_emissions_index_profile."""
    engine_uid = "01P06AL032"  # Engine for E45X (AE 3007A1E)

    # Cruise conditions
    air_pressure = np.array([19650])  # Units: Pa
    air_temperature = np.array([216])  # Units: K
    true_airspeed = np.array([247.5])  # Units: m/s
    fuel_flow_per_engine_descent = np.array([0.01])

    emissions = Emissions()
    edb_gaseous = emissions.edb_engine_gaseous[engine_uid]

    # Hydrocarbons (descent conditions with very low power)
    hc_ei = emissions_mod.hydrocarbon_emissions_index_ffm2(
        edb_gaseous,
        fuel_flow_per_engine_descent,
        true_airspeed,
        air_pressure,
        air_temperature,
    )
    np.testing.assert_almost_equal(hc_ei, np.array([0.0097947]), decimal=6)


def test_specific_humidity_correction_factor_display_45():
    """Test assertion following display (45) in Dubois & Paynter 2006."""

    # sea level values
    T = units.m_to_T_isa(0)
    p = units.m_to_pl(0) * 100
    rh = 0.6

    omega = ffm2._estimate_specific_humidity(T, p, rh)

    # value 0.00634 is taken from paragraph following (45)
    assert omega == pytest.approx(0.00634, abs=1e-5)


def test_example_page_11():
    """Confirm example on Dubois & Paynter 2006 page 11.

    The value for omega in the paper looks to be wrong.
    """
    # Given data
    T = 390 / 1.8
    p = units.ft_to_pl(39000) * 100
    rh = 0.6
    delta_amb = 0.1942
    theta_amb = 0.7519
    ei_sl_nox = 19.43
    ei_sl_co = 0.24

    omega = ffm2._estimate_specific_humidity(T, p, rh)
    # The value 0.00053 is taken from page 11 seems to be wrong ...
    # It should actually be 0.000053
    assert omega == pytest.approx(0.000053, abs=1e-5)

    q_factor = ffm2._get_humidity_correction_factor(omega)
    nox = ffm2.ei_at_cruise(ei_sl_nox, theta_amb, delta_amb, "NOX") * q_factor
    co = ffm2.ei_at_cruise(ei_sl_co, theta_amb, delta_amb, "CO")

    assert nox == pytest.approx(15.19, abs=1e-2)
    assert co == pytest.approx(0.50, abs=1e-2)


def test_emission_interp_all_nan():
    """Confirm emission functions gracefully handle missing data.

    Mock flight constructed below contains realworld data raising an exception in pycontrails.

    This test fails on pycontrails 0.20.0.
    """
    with pytest.warns(UserWarning, match="Time data is not np.datetime64"):
        fl = Flight(
            longitude=[-111.99457, -111.988735, -111.980604],
            latitude=[35.53877, 35.537842, 35.536582],
            altitude=[12496.8, 12496.8, 12496.8],
            time=["2022-02-01 03:00:21", "2022-02-01 03:00:23", "2022-02-01 03:00:25"],
            flight_id="killer whale",
        )
    fl["air_temperature"] = [np.nan, np.nan, np.nan]
    fl["specific_humidity"] = [0.002, 0.002, 0.003]
    fl["true_airspeed"] = [250.0, 240.0, 230.0]
    fl["fuel_flow"] = [0.5, 0.5, 0.6]

    fl.attrs["n_engine"] = 2
    fl.attrs["engine_uid"] = "01P06BR014"

    emissions = Emissions()

    out = emissions.eval(source=fl)
    assert np.isnan(out["nvpm_ei_n"]).all()
    assert np.isnan(out["nvpm_ei_m"]).all()

    fl.update(air_temperature=[230.0, 235.0, 240.0])
    out2 = emissions.eval(source=fl)

    assert np.isfinite(out2["nvpm_ei_n"]).all()
    assert np.isfinite(out2["nvpm_ei_m"]).all()


def test_t4_t2_ground_estimates():
    # Ground conditions from ICAO EDB (Airbus A320)
    true_airspeed = np.zeros(4)
    air_temperature = np.ones(4) * 0.5 * (275.8 + 287.2)
    air_pressure = np.ones(4) * 101325
    thrust_setting = np.array([0.134, 0.328, 0.873, 1.049]) / 1.049
    pressure_ratio = 27.1
    fuel = JetA()

    t4_t2_est = thrust_setting_nd(
        true_airspeed,
        thrust_setting,
        air_temperature,
        air_pressure,
        pressure_ratio,
        fuel.q_fuel,
        comp_efficiency=0.9,
        cruise=False,
    )

    t4_t2 = np.array([2.426554, 3.031491, 4.302515, 4.660068])
    assert t4_t2_est == pytest.approx(t4_t2, abs=1e-2)


def test_t4_t2_cruise_estimates():
    # Cruise conditions from ECLIF measurements (Airbus A320)
    fuel_flow = np.array([0.332192, 0.335372, 0.333815, 0.382108])
    air_temperature = np.array([218.116667, 217.845833, 217.720588, 233.272222])
    air_pressure = np.array([357.76667, 357.693750, 358.288235, 359.422222]) * 100
    true_airspeed = np.array([213.32261639, 212.55907557, 212.67744452, 233.40349122])
    mach_num = units.tas_to_mach_number(true_airspeed, air_temperature)

    theta_amb = jet.temperature_ratio(air_temperature)
    delta_amb = jet.pressure_ratio(air_pressure)
    fuel_flow_per_engine_est = jet.equivalent_fuel_flow_rate_at_sea_level(
        fuel_flow, theta_amb, delta_amb, mach_num
    )
    fuel_flow_per_engine_eclif = np.array([0.362298, 0.363894, 0.360877, 0.542128])
    assert fuel_flow_per_engine_est == pytest.approx(fuel_flow_per_engine_eclif, abs=1e-3)

    thrust_setting = fuel_flow_per_engine_est / 1.049
    pressure_ratio = 27.1
    fuel = JetA()

    t4_t2_est = thrust_setting_nd(
        true_airspeed,
        thrust_setting,
        air_temperature,
        air_pressure,
        pressure_ratio,
        fuel.q_fuel,
        comp_efficiency=0.9,
        cruise=True,
    )

    t4_t2_eclif = np.array([3.704481, 3.716919, 3.708952, 3.905757])
    assert t4_t2_est == pytest.approx(t4_t2_eclif, abs=1e-3)


def test_nvpm_ein_reductions_from_saf():
    """Test ``nvpm.nvpm_number_ei_pct_reduction_due_to_saf`` with some pinned values."""
    hydrogen_mass_content = np.array([14.11, 14.15, 14.1, 14.43, 14.47, 14.58, 14.4])
    thrust_setting = np.array([0.30, 0.65, 0.85, 0.3376, 0.4284, 0.3945, np.nan])
    d_nvpm_ein_pct = np.array([-25.55, -15.86, -7.23, -46.30, -42.34, -49.09, -37.96])
    d_nvpm_ein_pct_est = nvpm.nvpm_number_ei_pct_reduction_due_to_saf(
        hydrogen_mass_content, thrust_setting
    )
    np.testing.assert_allclose(d_nvpm_ein_pct_est, d_nvpm_ein_pct, atol=0.01)


@pytest.mark.parametrize("engine_uid", ["01P20CM128", "01P21GE216", "01P17GE210"])
def test_stage_combustors_data_length(engine_uid: str):
    """Confirm a few handpicked engine with multi-stage combustors have the correct data."""
    emissions = Emissions()
    edb_nvpm = emissions.edb_engine_nvpm[engine_uid]

    # Multistage combuster engines have 5 datapoints in the interpolation set
    assert len(edb_nvpm.nvpm_ei_m.fp) == 5
    assert len(edb_nvpm.nvpm_ei_m.xp) == 5
    assert len(edb_nvpm.nvpm_ei_n.fp) == 5
    assert len(edb_nvpm.nvpm_ei_n.xp) == 5

    # But the final three y values are all equal
    assert np.all(edb_nvpm.nvpm_ei_m.fp[2:] == edb_nvpm.nvpm_ei_m.fp[2])
    assert np.all(edb_nvpm.nvpm_ei_n.fp[2:] == edb_nvpm.nvpm_ei_n.fp[2])

    # The final x values are increasing
    assert np.all(np.diff(edb_nvpm.nvpm_ei_m.xp) > 0)
    assert np.all(np.diff(edb_nvpm.nvpm_ei_n.xp) > 0)



"""
Interface to ``pycontrails-bada`` extension.

Requires data files obtained with a
`BADA License <https://www.eurocontrol.int/model/bada>`_ from Eurocontrol.
"""

from __future__ import annotations

try:
    from pycontrails_bada import bada3, bada4, bada_model
    from pycontrails_bada.bada3 import BADA3
    from pycontrails_bada.bada4 import BADA4
    from pycontrails_bada.bada_interface import BADA
    from pycontrails_bada.bada_model import (
        BADAFlight,
        BADAFlightParams,
        BADAGrid,
        BADAGridParams,
        BADAParams,
    )

except ImportError as e:
    raise ImportError(
        'Failed to import `pycontrails-bada` extension. Install with `pip install "pycontrails-bada'
        ' @ git+ssh://git@github.com/contrailcirrus/pycontrails-bada.git"`'
    ) from e
else:
    __all__ = [
        "BADA",
        "BADA3",
        "BADA4",
        "BADAFlight",
        "BADAFlightParams",
        "BADAGrid",
        "BADAGridParams",
        "BADAParams",
        "bada3",
        "bada4",
        "bada_model",
    ]



"""Test `Spire` datalib."""

from __future__ import annotations

import pandas as pd
import pytest

from pycontrails.datalib import spire
from pycontrails.datalib.spire.spire import (
    _clean_trajectory_altitude,
    _separate_by_cruise_phase,
    _separate_by_on_ground,
)

from .conftest import get_static_path


def test_clean() -> None:
    """Test algorithms to identify and separate unique flight trajectories."""

    df = pd.read_parquet(get_static_path("flight-spire-data-cleaning.pq"))
    assert len(df.groupby(["icao_address", "tail_number", "aircraft_type_icao", "callsign"])) == 5

    clean = spire.clean(df)
    assert (
        len(clean.groupby(["icao_address", "tail_number", "aircraft_type_icao", "callsign"])) == 5
    )


def test_separate_using_ground_indicator():
    """Test algorithms to identify unique flight trajectories from on the ground indicator."""
    df = pd.read_parquet(get_static_path("flight-spire-data-cleaning.pq"))
    cdf = spire.clean(df)

    # Construct erroneous messages consisting of two unique flights with the same callsign
    test = cdf.loc[cdf["callsign"].isin(["SHT88J", "BAW506"])].copy()
    test["callsign"] = "killer-whale-1"

    # Unable to identify unique flights because metadata is the same
    assert len(test.groupby(["icao_address", "tail_number", "aircraft_type_icao", "callsign"])) == 1

    flight_ids = _separate_by_on_ground(test)
    assert len(flight_ids.unique()) == 2

    # Should handle slightly noisy "on_ground" signal
    # (there is already a single "on_ground" outlier near the end in the dataset)
    test.loc[2131:2132, "on_ground"] = False
    test.loc[4500:4502, "on_ground"] = True

    flight_ids = _separate_by_on_ground(test)
    assert len(flight_ids.unique()) == 2


def test_separate_with_multiple_cruise_phase():
    """Test algorithms to identify unique flight trajectories with multiple cruise phases."""
    df = pd.read_parquet(get_static_path("flight-spire-data-cleaning.pq"))
    cdf = spire.clean(df)

    # Construct erroneous messages consisting of two unique flights with the same callsign
    test = cdf.loc[cdf["callsign"].isin(["BAW506", "BAW507"])].copy()
    test["callsign"] = "killer-whale-2"

    # In `spire.identify_flights`, we need to clean up individual trajectory
    # altitudes before _separate_by_cruise_phase method works
    test = _clean_trajectory_altitude(test)

    # Unable to identify unique flights because metadata is the same
    assert len(test.groupby(["icao_address", "tail_number", "aircraft_type_icao", "callsign"])) == 1

    flight_ids = _separate_by_cruise_phase(test)
    assert len(flight_ids.unique()) == 1


@pytest.mark.xfail(reason="Test not complete")
def test_identify_flight_diversion() -> None:
    """Test algorithms to identify flight diversion, and no separation is done."""
    df = pd.read_parquet(get_static_path("flight-spire-data-cleaning.pq"))
    cdf = spire.clean(df)

    # Construct flight that is diverted
    test = cdf.loc[cdf["callsign"] == "BAW506"].copy()
    test.reset_index(drop=True, inplace=True)
    altitude_ft_adjusted = test["altitude_baro"].to_numpy()
    altitude_ft_adjusted[230:346] = test["altitude_baro"].values[539:655]
    altitude_ft_adjusted[346:462] = test["altitude_baro"].values[15:131]
    altitude_ft_adjusted[407:573] = 25000
    test["altitude_baro"] = altitude_ft_adjusted

    # Unable to identify unique flights because metadata is the same
    assert len(test.groupby(["icao_address", "tail_number", "aircraft_type_icao", "callsign"])) == 1

    flight_ids = _separate_by_cruise_phase(test)
    assert len(flight_ids.unique()) == 1



"""GFS Data Access.

References
----------
- `NOAA GFS <https://registry.opendata.aws/noaa-gfs-bdp-pds/>`_
- `Documentation <https://www.ncei.noaa.gov/products/weather-climate-models/global-forecast>`_
- `Parameter sets <https://www.nco.ncep.noaa.gov/pmb/products/gfs/>`_
"""

from __future__ import annotations

import hashlib
import logging
import pathlib
from contextlib import ExitStack
from datetime import datetime
from typing import TYPE_CHECKING, Any, Callable

import numpy as np
import pandas as pd
import xarray as xr
from overrides import overrides

import pycontrails
from pycontrails.core import cache, datalib, met
from pycontrails.datalib.gfs.variables import (
    PRESSURE_LEVEL_VARIABLES,
    SURFACE_VARIABLES,
    TOAUpwardLongwaveRadiation,
    TOAUpwardShortwaveRadiation,
    Visibility,
)
from pycontrails.utils.temp import temp_file
from pycontrails.utils.types import DatetimeLike

# optional imports
if TYPE_CHECKING:
    import botocore

LOG = logging.getLogger(__name__)

#: Default GFS AWS bucket
GFS_FORECAST_BUCKET = "noaa-gfs-bdp-pds"


class GFSForecast(datalib.MetDataSource):
    """GFS Forecast data access.

    Parameters
    ----------
    time : `datalib.TimeInput`
        The time range for data retrieval, either a single datetime or (start, end) datetime range.
        Input must be a single datetime-like or tuple of datetime-like (datetime,
        :class:`pandas.Timestamp`, :class:`numpy.datetime64`)
        specifying the (start, end) of the date range, inclusive.
        All times will be downloaded for a single forecast model run nearest to the start time
        (see :attr:`forecast_time`)
        If None, ``paths`` must be defined and all time coordinates will be loaded from files.
    variables : `datalib.VariableInput`
        Variable name (i.e. "temperature", ["temperature, relative_humidity"])
        See :attr:`pressure_level_variables` for the list of available variables.
    pressure_levels : `datalib.PressureLevelInput`, optional
        Pressure levels for data, in hPa (mbar)
        Set to [-1] for to download surface level parameters.
        Defaults to [-1].
    paths : str | list[str] | pathlib.Path | list[pathlib.Path] | None, optional
        Path to files to load manually.
        Can include glob patterns to load specific files.
        Defaults to None, which looks for files in the :attr:`cachestore` or GFS AWS bucket.
    grid : float, optional
        Specify latitude/longitude grid spacing in data.
        Defaults to 0.25.
    forecast_time : `DatetimeLike`, optional
        Specify forecast run by runtime.
        Defaults to None.
    cachestore : :class:`cache.CacheStore` | None, optional
        Cache data store for staging data files.
        Defaults to :class:`cache.DiskCacheStore`.
        If None, cachestore is turned off.
    show_progress : bool, optional
        Show progress when downloading files from GFS AWS Bucket.
        Defaults to False

    Examples
    --------
    >>> from datetime import datetime
    >>> from pycontrails.datalib.gfs import GFSForecast

    >>> # Store data files to local disk (default behavior)
    >>> times = ("2022-03-22 00:00:00", "2022-03-22 03:00:00", )
    >>> gfs = GFSForecast(times, variables="air_temperature", pressure_levels=[300, 250])
    >>> gfs
    GFSForecast
        Timesteps: ['2022-03-22 00', '2022-03-22 01', '2022-03-22 02', '2022-03-22 03']
        Variables: ['t']
        Pressure levels: [300, 250]
        Grid: 0.25
        Forecast time: 2022-03-22 00:00:00
        Steps: [0, 1, 2, 3]

    Notes
    -----
    - `NOAA GFS <https://registry.opendata.aws/noaa-gfs-bdp-pds/>`_
    - `Documentation <https://www.ncei.noaa.gov/products/weather-climate-models/global-forecast>`_
    - `Parameter sets <https://www.nco.ncep.noaa.gov/pmb/products/gfs/>`_
    - `GFS Documentation <https://www.emc.ncep.noaa.gov/emc/pages/numerical_forecast_systems/gfs/documentation.php>`_
    """  # noqa: E501

    __slots__ = ("client", "grid", "cachestore", "show_progress", "forecast_time")

    #: S3 client for accessing GFS bucket
    client: botocore.client.S3

    #: Lat / Lon grid spacing. One of [0.25, 0.5, 1]
    grid: float

    #: Show progress bar when downloading files from AWS
    show_progress: bool

    #: Base time of the previous GFS forecast based on input times
    forecast_time: datetime

    __marker = object()

    def __init__(
        self,
        time: datalib.TimeInput | None,
        variables: datalib.VariableInput,
        pressure_levels: datalib.PressureLevelInput = [-1],
        paths: str | list[str] | pathlib.Path | list[pathlib.Path] | None = None,
        grid: float = 0.25,
        forecast_time: DatetimeLike | None = None,
        cachestore: cache.CacheStore | None = __marker,  # type: ignore[assignment]
        show_progress: bool = False,
    ):
        try:
            import boto3
            import botocore
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                "`gfs` module dependencies are missing. "
                "Please install all required dependencies using `pip install -e .[gfs]`"
            ) from e

        # inputs
        self.paths = paths
        if cachestore is self.__marker:
            cachestore = cache.DiskCacheStore()
        self.cachestore = cachestore
        self.show_progress = show_progress

        if time is None and paths is None:
            raise ValueError("Time input is required when paths is None")

        self.timesteps = datalib.parse_timesteps(time, freq="1H")
        self.pressure_levels = datalib.parse_pressure_levels(
            pressure_levels, self.supported_pressure_levels
        )
        self.variables = datalib.parse_variables(variables, self.supported_variables)
        self.grid = datalib.parse_grid(grid, [0.25, 0.5, 1])

        # note GFS allows unsigned requests (no credentials)
        # https://stackoverflow.com/questions/34865927/can-i-use-boto3-anonymously/34866092#34866092
        self.client = boto3.client(
            "s3", config=botocore.client.Config(signature_version=botocore.UNSIGNED)
        )

        # set specific forecast time is requested
        if forecast_time is not None:
            forecast_time_pd = pd.to_datetime(forecast_time)
            if forecast_time_pd.hour not in [0, 6, 12, 18]:
                raise ValueError("Forecast hour must be on one of 00, 06, 12, 18")

            self.forecast_time = datalib.round_hour(forecast_time_pd.to_pydatetime(), 6)

        # if no specific forecast is requested, set the forecast time using timesteps
        elif self.timesteps:
            # round first element to the nearest 6 hour time (00, 06, 12, 18 UTC) for forecast_time
            self.forecast_time = datalib.round_hour(self.timesteps[0], 6)

        # when no forecast_time or time input, forecast_time is defined in _open_and_cache

    def __repr__(self) -> str:
        base = super().__repr__()
        return (
            f"{base}\n\tForecast time: {getattr(self, 'forecast_time', '')}\n\tSteps:"
            f" {getattr(self, 'steps', '')}"
        )

    @property
    def supported_pressure_levels(self) -> list[int]:
        """Get pressure levels available.

        Returns
        -------
        list[int]
            List of integer pressure level values
        """
        return [
            1000,
            975,
            950,
            925,
            900,
            850,
            800,
            750,
            700,
            650,
            600,
            550,
            500,
            450,
            400,
            350,
            300,
            250,
            200,
            150,
            100,
            70,
            50,
            40,
            30,
            20,
            15,
            10,
            7,
            5,
            3,
            2,
            1,
            -1,
        ]

    @property
    def pressure_level_variables(self) -> list[met.MetVariable]:
        """GFS pressure level parameters.

        Returns
        -------
        list[MetVariable] | None
            List of MetVariable available in datasource
        """
        return PRESSURE_LEVEL_VARIABLES

    @property
    def single_level_variables(self) -> list[met.MetVariable]:
        """GFS surface level parameters.

        Returns
        -------
        list[MetVariable] | None
            List of MetVariable available in datasource
        """
        return SURFACE_VARIABLES

    @property
    def hash(self) -> str:
        """Generate a unique hash for this datasource.

        Returns
        -------
        str
            Unique hash for met instance (sha1)
        """
        hashstr = (
            f"{self.__class__.__name__}{self.timesteps}{self.variable_shortnames}"
            f"{self.pressure_levels}{self.grid}{self.forecast_time}"
        )
        return hashlib.sha1(bytes(hashstr, "utf-8")).hexdigest()

    @property
    def step_offset(self) -> int:
        """Difference between :attr:`forecast_time` and first timestep.

        Returns
        -------
        int
            Number of steps to offset in order to retrieve data starting from input time.
            Returns 0 if :attr:`timesteps` is empty when loading from :attr:`paths`.
        """
        if self.timesteps:
            return int((self.timesteps[0] - self.forecast_time).total_seconds() // 3600)

        return 0

    @property
    def steps(self) -> list[int]:
        """Forecast steps from :attr:`forecast_time` corresponding within input :attr:`time`.

        Returns
        -------
        list[int]
            List of forecast steps relative to :attr:`forecast_time`
        """
        return [self.step_offset + i for i in range(len(self.timesteps))]

    @property
    def _grid_string(self) -> str:
        """Return filename string for grid spacing."""
        if self.grid == 0.25:
            return "0p25"
        elif self.grid == 0.5:
            return "0p50"
        elif self.grid == 1:
            return "1p00"
        else:
            raise ValueError(
                f"Unsupported grid spacing {self.grid}. Must be one of 0.25, 0.5, or 1.0."
            )

    @property
    def forecast_path(self) -> str:
        """Construct forecast path in bucket for :attr:`forecast_time`.

        String template:

            GFS_FORECAST_BUCKET/gfs.YYYYMMDD/HH/atmos/{filename}",

        Returns
        -------
        str
            Bucket prefix for forecast files.
        """
        datestr = self.forecast_time.strftime("%Y%m%d")
        forecast_hour = str(self.forecast_time.hour).zfill(2)
        return f"gfs.{datestr}/{forecast_hour}/atmos"

    def filename(self, step: int) -> str:
        """Construct grib filename to retrieve from GFS bucket.

        String template:

            gfs.tCCz.pgrb2.GGGG.fFFF

        - ``CC`` is the model cycle runtime (i.e. 00, 06, 12, 18)
        - ``GGGG`` is the grid spacing
        - ``FFF`` is the forecast hour of product from 000 - 384

        Parameters
        ----------
        step : int
            Integer step relative to forecast time

        Returns
        -------
        str
            Forecast filenames to retrieve from GFS bucket.

        References
        ----------
        - https://www.nco.ncep.noaa.gov/pmb/products/gfs/
        """
        forecast_hour = str(self.forecast_time.hour).zfill(2)
        return f"gfs.t{forecast_hour}z.pgrb2.{self._grid_string}.f{str(step).zfill(3)}"

    @overrides
    def create_cachepath(self, t: datetime) -> str:
        if self.cachestore is None:
            raise ValueError("self.cachestore attribute must be defined to create cache path")

        # get forecast_time and step for specific file
        datestr = self.forecast_time.strftime("%Y%m%d-%H")

        # get step relative to forecast forecast_time
        step = self.step_offset + self.timesteps.index(t)

        # single level or pressure level
        suffix = f"gfs{'sl' if self.pressure_levels == [-1] else 'pl'}{self.grid}"

        # return cache path
        return self.cachestore.path(f"{datestr}-{step}-{suffix}.nc")

    @overrides
    def download_dataset(self, times: list[datetime]) -> None:
        # get step relative to forecast forecast_time
        LOG.debug(
            f"Downloading GFS forecast for forecast time {self.forecast_time} and timesteps {times}"
        )

        # download grib file for each step file
        for t in times:
            self._download_file(t)

    @overrides
    def cache_dataset(self, dataset: xr.Dataset) -> None:
        # if self.cachestore is None:
        #     LOG.debug("Cache is turned off, skipping")
        #     return

        raise NotImplementedError("GFS caching only implemented with download")

    @overrides
    def open_metdataset(
        self,
        dataset: xr.Dataset | None = None,
        xr_kwargs: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> met.MetDataset:
        xr_kwargs = xr_kwargs or {}

        #  short-circuit file paths if provided
        if dataset is not None:
            raise NotImplementedError("GFS data source does not support passing local dataset")

        elif self.paths is not None:
            raise NotImplementedError("GFS data source does not support passing local paths")

            # TODO: This should work but i have type issues

            # if isinstance(self.paths, (str, pathlib.Path)):
            #     self.paths: list[str] | list[pathlib.Path] = [self.paths]

            # for (filepath, t) in zip(self.paths, self.timesteps):
            #     self._open_gfs_dataset(filepath, t)

        # load from cache or download
        else:
            if self.cachestore is None:
                raise ValueError("Cachestore is required to download data")

            # confirm files are downloaded any remote (AWS, of Cache)
            self.download(**xr_kwargs)

            # ensure all files are guaranteed to be available locally here
            # this would download a file from a remote (e.g. GCP) cache
            disk_cachepaths = [self.cachestore.get(f) for f in self._cachepaths]

            # run MetDataset constructor
            ds = self.open_dataset(disk_cachepaths, **xr_kwargs)

            # TODO: corner case
            # If any files are already cached, they will not have the version attached
            if "pycontrails_version" not in ds.attrs:
                ds.attrs["pycontrails_version"] = pycontrails.__version__

        # run the same GFS-specific processing on the dataset
        mds = self._process_dataset(ds, **kwargs)

        # set TOAUpwardShortwaveRadiation, TOAUpwardLongwaveRadiation step 0 == step 1
        for key in [
            TOAUpwardShortwaveRadiation.standard_name,
            TOAUpwardLongwaveRadiation.standard_name,
        ]:
            # if step 0 (forecast time) exists in dimension
            forecast_time = mds.data["forecast_time"].values
            if key in mds.data and forecast_time in mds.data["time"]:
                # make sure this isn't the only time in the dataset
                if np.all(mds.data["time"].values == forecast_time):
                    raise RuntimeError(
                        f"GFS datasets with data variable {key} must have at least one timestep"
                        f" after the forecast time to estimate the value of {key} at step 0"
                    )

                # set the 0th value of the data to the 1st value
                # TODO: this assumption may not be universally applicable!
                mds.data[key][dict(time=0)] = mds.data[key][dict(time=1)]

        return mds

    def _download_file(self, t: datetime) -> None:
        """Download data file for forecast time and step.

        Overwrites files if they already exists.

        Parameters
        ----------
        t : datetime
            Timestep to download

        Notes
        -----
        - ``f000``:
          https://www.nco.ncep.noaa.gov/pmb/products/gfs/gfs.t00z.pgrb2.0p25.f000.shtml
        - ``f000 - f384``:
          https://www.nco.ncep.noaa.gov/pmb/products/gfs/gfs.t00z.pgrb2.0p25.f003.shtml
        """

        if self.cachestore is None:
            raise ValueError("Cachestore is required to download data")

        # construct filenames for each file
        step = self.step_offset + self.timesteps.index(t)
        filename = self.filename(step)
        aws_key = f"{self.forecast_path}/{filename}"

        # Open ExitStack to control temp_file context manager
        with ExitStack() as stack:
            # hold downloaded file in named temp file
            temp_grib_filename = stack.enter_context(temp_file())

            # retrieve data from AWS S3
            LOG.debug(f"Downloading GFS file {filename} from AWS bucket to {temp_grib_filename}")
            if self.show_progress:
                _download_with_progress(
                    self.client, GFS_FORECAST_BUCKET, aws_key, temp_grib_filename, filename
                )
            else:
                self.client.download_file(
                    Bucket=GFS_FORECAST_BUCKET, Key=aws_key, Filename=temp_grib_filename
                )

            ds = self._open_gfs_dataset(temp_grib_filename, t)

            # write out data to temp, close grib file
            temp_nc_filename = stack.enter_context(temp_file())
            ds.to_netcdf(path=temp_nc_filename, mode="w")

            # put each hourly file into cache
            self.cachestore.put(temp_nc_filename, self.create_cachepath(t))

    def _open_gfs_dataset(self, filepath: str | pathlib.Path, t: datetime) -> xr.Dataset:
        """Open GFS grib file for one forecast timestep.

        Parameters
        ----------
        filepath : str | pathlib.Path
            Path to GFS forecast file
        t : datetime
            Timestep corresponding with GFS forecast

        Returns
        -------
        xr.Dataset
            GFS dataset
        """
        # translate into netcdf from grib
        LOG.debug(f"Translating {filepath} for timestep {str(t)} into netcdf")

        # get step for timestep
        step = self.step_offset + self.timesteps.index(t)

        # open file for each variable short name individually
        ds = xr.Dataset()
        for variable in self.variables:
            # radiation data is not available in the 0th step
            if step == 0 and variable in [
                TOAUpwardShortwaveRadiation,
                TOAUpwardLongwaveRadiation,
            ]:
                LOG.debug(
                    "Radiation data is not provided for the 0th step in GFS. Setting to np.nan"
                    " using Visibility variable"
                )
                v = Visibility
            else:
                v = variable

            tmpds = xr.open_dataset(
                filepath,
                filter_by_keys={"typeOfLevel": v.level_type, "shortName": v.short_name},
                engine="cfgrib",
            )

            if not len(ds):
                ds = tmpds
            else:
                ds[v.short_name] = tmpds[v.short_name]

            # set all radiation data to np.nan in the 0th step
            if step == 0 and variable in [
                TOAUpwardShortwaveRadiation,
                TOAUpwardLongwaveRadiation,
            ]:
                ds = ds.rename({Visibility.short_name: variable.short_name})
                ds[variable.short_name] = np.nan

        # for pressure levels, need to rename "level" field and downselect
        if self.pressure_levels != [-1]:
            ds = ds.rename({"isobaricInhPa": "level"})
            ds = ds.sel(dict(level=self.pressure_levels))

        # for single level, and singular pressure levels, add the level dimension
        if len(self.pressure_levels) == 1:
            ds = ds.expand_dims({"level": self.pressure_levels})

        # rename fields and swap time dimension for step
        ds = ds.rename({"time": "forecast_time"})
        ds = ds.rename({"valid_time": "time"})
        ds = ds.expand_dims("time")

        # drop step/number
        ds = ds.drop_vars(["step", "nominalTop", "surface"], errors="ignore")

        return ds

    def _process_dataset(self, ds: xr.Dataset, **kwargs: Any) -> met.MetDataset:
        """Process the :class:`xr.Dataset` opened from cache or local files.

        Parameters
        ----------
        ds : xr.Dataset
            Dataset loaded from netcdf cache files or input paths.
        **kwargs : Any
            Keyword arguments passed through directly into :class:`MetDataset` constructor.

        Returns
        -------
        MetDataset
        """

        # downselect dataset if only a subset of times, pressure levels, or variables are requested
        ds = ds[self.variable_shortnames]

        if self.timesteps:
            ds = ds.sel(time=self.timesteps)
        else:
            # set timesteps from dataset "time" coordinates
            # np.datetime64 doesn't covert to list[datetime] unless its unit is us
            self.timesteps = ds["time"].values.astype("datetime64[us]").tolist()

        # if "level" is not in dims and
        # length of the requested pressure levels is 1
        # expand the dims with this level
        if "level" not in ds.dims and len(self.pressure_levels) == 1:
            ds = ds.expand_dims({"level": self.pressure_levels})

        else:
            ds = ds.sel(dict(level=self.pressure_levels))

        # harmonize variable names
        ds = met.standardize_variables(ds, self.variables)

        if "cachestore" not in kwargs:
            kwargs["cachestore"] = self.cachestore

        ds.attrs["met_source"] = type(self).__name__
        return met.MetDataset(ds, **kwargs)


def _download_with_progress(
    client: botocore.client.S3, bucket: str, key: str, filename: str, label: str
) -> None:
    """Download with `tqdm` progress bar.

    Parameters
    ----------
    client : botocore.client.S3
        S3 Client
    bucket : str
        AWS Bucket
    key : str
        Key within bucket to download
    filename : str
        Local filename to download to
    label : str
        Progress label

    Raises
    ------
    ModuleNotFoundError
        Raises if tqdm can't be found
    """

    try:
        from tqdm import tqdm
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(
            "Download with progress requires the `tqdm` module, "
            "which can be installed using `pip install pycontrails[gfs]`. "
            "Alternatively, set instance attribute `show_progress=False`."
        ) from e

    meta = client.head_object(Bucket=bucket, Key=key)
    filesize = meta["ContentLength"]

    def hook(t: Any) -> Callable:
        def inner(bytes_amount: Any) -> None:
            t.update(bytes_amount)

        return inner

    with tqdm(total=filesize, unit="B", unit_scale=True, desc=label) as t:
        client.download_file(Bucket=bucket, Key=key, Filename=filename, Callback=hook(t))
